#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Epics module. See __init__.py for more detailed documentation
"""
__all__ = ['EpicsAuthorityNameValidator', 'EpicsDeviceNameValidator',
           'EpicsAttributeNameValidator']

import re
from taurus.core.taurusvalidator import (TaurusAttributeNameValidator,
                                         TaurusDeviceNameValidator,
                                         TaurusAuthorityNameValidator)

# legal chars on PV names: a-z A-Z 0-9 _ - : . [ ] < > ;
# ... but "[" "]" and "." are discouraged:
# http://www.aps.anl.gov/epics/tech-talk/2006/msg00431.php
PV_CHARS = r'[a-zA-Z0-9_\-:;\<\>' + r'[\.\[\]]'


class EpicsAuthorityNameValidator(TaurusAuthorityNameValidator):
    """Validator for Epics authority names. For now, the only supported
    authority is "//":
    """
    scheme = '(ca|epics)'
    authority = '//'
    path = '(?!)'
    query = '(?!)'
    fragment = '(?!)'

    def getNames(self, fullname, factory=None):
        if self.isValid(fullname):
            return 'ca://', '//', ''
        return None


class EpicsDeviceNameValidator(TaurusDeviceNameValidator):
    """Validator for Epics device names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - devname: device name (only empty string allowed for now)

    Note: brackets on the group name indicate that this group will only contain
    a string if the URI contains it.
    """

    scheme = '(ca|epics)'
    authority = EpicsAuthorityNameValidator.authority
    path = r'/(?P<devname>)'  # (only empty string allowed for now)
    query = '(?!)'
    fragment = '(?!)'

    def getNames(self, fullname, factory=None):
        if self.isValid(fullname):
            return 'ca:', '', ''
        return None


class EpicsAttributeNameValidator(TaurusAttributeNameValidator):
    """Validator for Epics attribute names. Apart from the standard named
    groups (scheme, authority, path, query and fragment), the following named
    groups are created:

     - attrname: attribute name (an epics PV name).

    Note: brackets on the group name indicate that this group will only contain
    a value if the URI contains it.
    """
    scheme = '(ca|epics)'
    authority = EpicsAuthorityNameValidator.authority
    path = r'(?P<attrname>%s+?(\.(?P<_field>[A-Z]+))?)' % PV_CHARS
    query = '(?!)'
    fragment = '[^# ]*'

    def getNames(self, fullname, factory=None, fragment=False):
        """reimplemented from :class:`TaurusDeviceNameValidator`"""

        groups = self.getUriGroups(fullname)
        if groups is None:
            return None

        complete = 'ca:%s' % groups['attrname']
        normal = groups['attrname']
        short = normal

        # return fragment if requested
        if fragment:
            key = groups.get('fragment', None)
            return complete, normal, short, key
        return complete, normal, short



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""Taurus import tests"""
from __future__ import absolute_import

import sys
import unittest


class TaurusImportTestCase(unittest.TestCase):

    """
    Test if all the submodules can be imported
    """

    def setUp(self):
        """Preconditions: moduleexplorer utility has to be available """
        from .moduleexplorer import ModuleExplorer
        self.explore = ModuleExplorer.explore

    def testImportSubmodules(self):
        """
        Check that all taurus submodules import without problems

        Expected Results: It is expected to get no warning message
        on module importing
        """
        exclude_patterns = [r'taurus\.qt\.qtgui\.extra_.*',
                            r'taurus\.qt\.qtgui\.qwt5',
                            r'taurus\.external\.qt\.QtUiTools',
                            r'taurus\.external\.qt\.QtWebKit',
                            r'taurus\.external\.qt\.Qwt5',
                            ]

        try:
            import PyTango
        except ImportError:
            exclude_patterns.append(r'taurus\.core\.tango')
        try:
            import epics
        except ImportError:
            exclude_patterns.append(r'taurus\.core\.epics')

        from taurus.external.qt import PYSIDE2
        if PYSIDE2:
            exclude_patterns.append(r'taurus\.external\.qt\.QtDesigner')
            exclude_patterns.append(r'taurus\.qt\.qtdesigner')

        moduleinfo, wrn = self.explore('taurus', verbose=False,
                                       exclude_patterns=exclude_patterns)
        msg = None
        if wrn:
            msg = '\n%s' % '\n'.join(list(zip(*wrn))[1])
        self.assertEqual(len(wrn), 0, msg=msg)


if __name__ == "__main__":
    unittest.main()



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides the taurus Command Line Interface common options.

It is based on the click module to provide commonly used flags/options.
They are used by taurus commands and can be used by plugin to easily
extend its functionality.

Example 1: Add custom subcommand
script::

    import click

    import taurus.cli.common


    @click.command("bar")
    @taurus.cli.common.poll_period
    @taurus.cli.common.default_formatter
    @taurus.cli.common.window_name("Super Bar")
    def bar(poll_period, default_formatter, window_name):
        ...


    if __name__ == '__main__':
        bar()

Example 2: Add custom subcommands' group
script::

    import click

    import taurus.cli.common


    @click.group("foo")
    def foo():
        pass


    @foo.command('cmd1')
    @taurus.cli.common.models
    @taurus.cli.common.config_file
    @taurus.cli.common.window_name("Super Foo (cmd1)")
    def cmd1(models, config_file, window_name):
        ...


    @foo.command('trend')
    @taurus.cli.common.model
    @taurus.cli.common.serial_mode
    @taurus.cli.common.poll_period
    @taurus.cli.common.default_formatter
    def cmd2(model, serial_mode, poll_period, default_formatter):
        ...


    if __name__ == '__main__':
        foo()

"""

import click


__levels = ["Critical", "Error", "Warning", "Info", "Debug", "Trace"]
try:
    __log_choice = click.Choice(__levels, case_sensitive=False)
except TypeError:  # click v< 7 does not allow case_sensitive option
    __log_choice = click.Choice(__levels)


log_level = click.option(
    "--log-level",
    "log_level",
    type=__log_choice,
    default="Info",
    show_default=True,
    help="Show only logs with priority LEVEL or above",
)

config_file = click.option(
    "--config",
    "config_file",
    type=click.File("rb"),
    help="Configuration file for initialization",
)


def window_name(default):
    window_name = click.option(
        "--window-name",
        "window_name",
        default=default,
        help="Name of the window",
    )
    return window_name


models = click.argument("models", nargs=-1,)

demo = click.option("--demo", is_flag=True, help="Show a demo of the widget",)

model = click.argument("model", nargs=1, required=False,)

poll_period = click.option(
    "--polling-period",
    "polling_period",
    type=click.INT,
    metavar="MILLISEC",
    default=None,
    help="Change the Default Taurus polling period",
)

serial_mode = click.option(
    "--serialization-mode",
    "serialization_mode",
    type=click.Choice(["Serial", "Concurrent", "TangoSerial"]),
    default=None,
    show_default=True,
    help=(
        "Set the default Taurus serialization mode for those "
        + "models that do not explicitly define it)"
    ),
)

default_formatter = click.option(
    "--default-formatter",
    "default_formatter",
    type=click.STRING,
    metavar="FORMATTER",
    default=None,
    help="Override the default formatter (use with caution!)",
)

list_alternatives = click.option(
    "--ls-alt",
    is_flag=True,
    help="List the available alternative implementations",
)

use_alternative = click.option(
    "--use-alt",
    metavar="ALT",
    type=click.STRING,
    default=None,
    help="Use ALT alternative implementation",
)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Extension of :mod:`guiqwt.image`
"""
__all__ = ["TaurusImageItem", "TaurusRGBImageItem", "TaurusTrend2DItem",
           "TaurusTrend2DScanItem", "TaurusEncodedImageItem",
           "TaurusEncodedRGBImageItem"]

from taurus.core.units import Quantity
from taurus.external.qt import Qt
from taurus.qt.qtgui.base import TaurusBaseComponent
from taurus.qt.qtcore.util import baseSignal
import taurus.core
from taurus.core.util.containers import ArrayBuffer

from guiqwt.image import ImageItem, RGBImageItem, XYImageItem
from guiqwt.image import INTERP_NEAREST, INTERP_LINEAR

import numpy


class TaurusBaseImageItem(TaurusBaseComponent):
    '''A ImageItem that gets its data from a taurus attribute'''

    dataChanged = baseSignal('dataChanged')

    def setModel(self, model):
        # do the standard stuff
        TaurusBaseComponent.setModel(self, model)
        #... and fire a fake event for initialization
        try:
            value = self.getModelObj().read()
            self.fireEvent(
                self, taurus.core.taurusbasetypes.TaurusEventType.Change, value)
        except:
            pass

    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_value is None or getattr(evt_value, 'rvalue', None) is None:
            self.debug('Ignoring event from %s' % repr(evt_src))
            return
        v = evt_value.rvalue
        if isinstance(v, Quantity):
            v = v.magnitude
            # TODO: units should be used for setting some title in the colorbar
        try:
            v = self.filterData(v)
        except Exception as e:
            self.info('Ignoring event. Reason: %s', e.message)
            return
        # this is the range of the z axis (color scale)
        lut_range = self.get_lut_range()
        # if the range was not set, make it None (autoscale z axis)
        if lut_range[0] == lut_range[1]:
            lut_range = None
        self.set_data(v, lut_range=lut_range)
        self.dataChanged.emit()
        p = self.plot()

        if p is not None:
            p.update_colormap_axis(self)
            p.replot()

    def filterData(self, data):
        '''Reimplement this method if you want to pre-process
        the data that will be passed to set_data.

        It should return something acceptable by :meth:`setData`
        and raise an exception if the data cannot be processed.

        This default implementation casts array types not
        supported by guiqwt to numpy.int32

        See:
          - http://code.google.com/p/guiqwt/issues/detail?id=44 and
          - https://sourceforge.net/tracker/?func=detail&atid=484769&aid=3603991&group_id=57612
          - https://sourceforge.net/p/tauruslib/tickets/33/
        '''
        try:
            dtype = data.dtype
            v = data
        except:
            v = numpy.array(data)  # note that this is potentially expensive
            dtype = v.dtype

        if dtype not in (float, numpy.double, numpy.int32, numpy.uint16,
                         numpy.int16, numpy.uint8, numpy.int8, bool):
            # note: numpy.uint32 was not included because of
            # https://sourceforge.net/p/tauruslib/tickets/33/
            try:
                self.debug('casting to numpy.int32')
                v = numpy.int32(v)
            except OverflowError:
                raise OverflowError(
                    "type %s not supported by guiqwt and cannot be casted to int32" % repr(v.dtype))

        return v


class TaurusEncodedBaseImageItem(TaurusBaseImageItem):
    '''A ImageItem that gets its data from a taurus DevEncoded attribute'''

    def setModel(self, model):
        # do the standard stuff
        TaurusBaseComponent.setModel(self, model)
        #... and fire a fake event for initialization
        try:
            fmt, value = self.codec.decode(self.getModelObj().read())
            self.fireEvent(
                self, taurus.core.taurusbasetypes.TaurusEventType.Change, value)
        except:
            pass

    def filterData(self, data):
        '''reimplementation to decode data using the DevEncoded codecs'''
        if type(data) == tuple:
            from taurus.core.util.codecs import CodecFactory
            codec = CodecFactory().getCodec(data[0])

            try:
                fmt, decoded_data = codec.decode(data)
            except Exception as e:
                self.info('Decoder error: %s', e.message)
                raise e

            try:
                dtype = decoded_data.dtype
                v = decoded_data
            except:
                # note that this is potentially expensive
                v = numpy.array(decoded_data)
                dtype = v.dtype

            if dtype not in (float, numpy.double, numpy.int32, numpy.uint16,
                             numpy.int16, numpy.uint8, numpy.int8, bool):
                # note: numpy.uint32 was not included because of
                # https://sourceforge.net/p/tauruslib/tickets/33/
                try:
                    self.debug('casting to numpy.int32')
                    v = numpy.int32(v)
                except OverflowError:
                    raise OverflowError(
                        "type %s not supported by guiqwt and cannot be casted to int32" % repr(v.dtype))

            return v
        else:
            raise ValueError(
                'Unexpected data type (%s) for DevEncoded attribute (tuple expected)' % type(data))


class TaurusImageItem(ImageItem, TaurusBaseImageItem):
    '''A ImageItem that gets its data from a taurus attribute'''

    def __init__(self, param=None):
        ImageItem.__init__(self, numpy.zeros((1, 1)), param=param)
        TaurusBaseImageItem.__init__(self, self.__class__.__name__)


class TaurusEncodedImageItem(ImageItem, TaurusEncodedBaseImageItem):
    '''A ImageItem that gets its data from a DevEncoded attribute'''

    def __init__(self, param=None):
        ImageItem.__init__(self, numpy.zeros((1, 1)), param=param)
        TaurusEncodedBaseImageItem.__init__(self, self.__class__.__name__)


class TaurusXYImageItem(XYImageItem, TaurusBaseImageItem):
    '''A XYImageItem that gets its data from a taurus attribute'''

    def __init__(self, param=None):
        XYImageItem.__init__(self, numpy.arange(2), numpy.arange(
            2), numpy.zeros((2, 2)), param=param)
        TaurusBaseImageItem.__init__(self, self.__class__.__name__)


class TaurusRGBImageItem(RGBImageItem, TaurusBaseImageItem):
    '''A RGBImageItem that gets its data from a taurus attribute'''

    def __init__(self, param=None):
        RGBImageItem.__init__(self, numpy.zeros((1, 1, 3)), param=param)
        TaurusBaseImageItem.__init__(self, self.__class__.__name__)

    def set_data(self, data, lut_range=None, **kwargs):
        '''dummy reimplementation to accept the lut_range kwarg (just ignoring it)'''
        return RGBImageItem.set_data(self, data, **kwargs)


class TaurusEncodedRGBImageItem(RGBImageItem, TaurusEncodedBaseImageItem):
    '''A RGBImageItem that gets its data from a DevEncoded attribute'''

    def __init__(self, param=None):
        RGBImageItem.__init__(self, numpy.zeros((1, 1, 3)), param=param)
        TaurusEncodedBaseImageItem.__init__(self, self.__class__.__name__)

    def set_data(self, data, lut_range=None, **kwargs):
        '''dummy reimplementation to accept the lut_range kwarg (just ignoring it)'''
        return RGBImageItem.set_data(self, data, **kwargs)


class TaurusTrend2DItem(XYImageItem, TaurusBaseComponent):
    """
    A XYImageItem that is constructed by stacking 1D arrays from events from 
    a Taurus 1D attribute
    """

    scrollRequested = baseSignal('scrollRequested', object, object, object)
    dataChanged = baseSignal('dataChanged')

    def __init__(self, param=None, buffersize=512, stackMode='datetime'):
        """
        :param param: param to be passed to XYImageItem constructor
        :param buffersize: (int) size of the stack
        :param stackMode: (str) can be 'datetime', 'timedelta' or 'event'
        """
        XYImageItem.__init__(self, numpy.arange(2), numpy.arange(
            2), numpy.zeros((2, 2)), param=param)
        TaurusBaseComponent.__init__(self, self.__class__.__name__)
        self.maxBufferSize = buffersize
        self._yValues = None
        self._xBuffer = None
        self._zBuffer = None
        self.stackMode = stackMode
        self.set_interpolation(INTERP_NEAREST)
        self.__timeOffset = None

        # Config properties
        self.registerConfigProperty(self.get_lut_range,
                                    self.set_lut_range,
                                    'lut_range'
                                    )
        self.registerConfigProperty(self._get_interpolation_cfg,
                                    self._set_interpolation_cfg,
                                    'interpolation'
                                    )
        self.registerConfigProperty(self.get_color_map_name,
                                    self.set_color_map,
                                    'color_map'
                                    )

    def _get_interpolation_cfg(self):
        ret = self.get_interpolation()
        if len(ret) == 2:
            ret = (ret[0], len(ret[1]))
        return ret

    def _set_interpolation_cfg(self, interpolate_cfg):
        self.set_interpolation(*interpolate_cfg)

    def setBufferSize(self, buffersize):
        '''sets the size of the stack

        :param buffersize: (int) size of the stack
        '''
        self.maxBufferSize = buffersize
        try:
            if self._xBuffer is not None:
                self._xBuffer.setMaxSize(buffersize)
            if self._zBuffer is not None:
                self._zBuffer.setMaxSize(buffersize)
        except ValueError:
            self.info(
                'buffer downsizing  requested. Current contents will be discarded')
            self._xBuffer = None
            self._zBuffer = None

    def setModel(self, model):
        # do the standard stuff
        TaurusBaseComponent.setModel(self, model)
        #... and fire a fake event for initialization
        try:
            value = self.getModelObj().read()
            self.fireEvent(
                self, taurus.core.taurusbasetypes.TaurusEventType.Change, value)
        except:
            pass

    def handleEvent(self, evt_src, evt_type, evt_value):
        if evt_value is None or getattr(evt_value, 'rvalue', None) is None:
            self.debug('Ignoring event from %s' % repr(evt_src))
            return
        plot = self.plot()
        if plot is None:
            return

        # initialization
        ySize = len(evt_value.rvalue)
        if self._yValues is None:
            self._yValues = numpy.arange(ySize, dtype='d')
        if self._xBuffer is None:
            self._xBuffer = ArrayBuffer(numpy.zeros(
                min(128, self.maxBufferSize), dtype='d'), maxSize=self.maxBufferSize)
        if self._zBuffer is None:
            self._zBuffer = ArrayBuffer(numpy.zeros(
                (min(128, self.maxBufferSize), ySize), dtype='d'), maxSize=self.maxBufferSize)
            return

        # check that new data is compatible with previous data
        if ySize != self._yValues.size:
            self.info('Incompatible shape in data from event (orig=%i, current=%i). Ignoring' % (
                self._yValues.size, ySize))
            return

        # update x values
        if self.stackMode == 'datetime':
            x = evt_value.time.totime()
            if self.__timeOffset is None:
                self.__timeOffset = x
                plot.set_axis_title('bottom', 'Time')
                plot.set_axis_unit('bottom', '')

        elif self.stackMode == 'deltatime':
            try:
                x = evt_value.time.totime() - self.__timeOffset
            except TypeError:  # this will happen if self.__timeOffset has not been initialized
                self.__timeOffset = evt_value.time.totime()
                x = 0
                plot.set_axis_title('bottom', 'Time since %s' %
                                    evt_value.time.isoformat())
                plot.set_axis_unit('bottom', '')
        elif self.stackMode == 'event':
            try:
                # +numpy.random.randint(0,4) #for debugging we can put a variable step
                step = 1
                x = self._xBuffer[-1] + step
            except IndexError:  # this will happen when the x buffer is empty
                x = 0
                plot.set_axis_title('bottom', 'Event #')
                plot.set_axis_unit('bottom', '')
        else:
            raise ValueError('Unsupported stack mode %s' % self.stackMode)

        if len(self._xBuffer) and x <= self._xBuffer[-1]:
            self.info('Ignoring event (non-increasing x value)')
            return
        self._xBuffer.append(x)

        # update z
        rvalue = evt_value.rvalue
        if isinstance(evt_value.rvalue, Quantity):
            rvalue = evt_value.rvalue.magnitude
            # TODO: units should be checked for coherence with previous values
        self._zBuffer.append(rvalue)

        # check if there is enough data to start plotting
        if len(self._xBuffer) < 2:
            self.info('waiting for at least 2 values to start plotting')
            return

        x = self._xBuffer.contents()
        y = self._yValues
        z = self._zBuffer.contents().transpose()

        # Use previous LUT range (z axis range), or set to None (autoscale)
        # if it is uninitialized
        lut_range = self.get_lut_range()
        if lut_range[0] == lut_range[1]:
            lut_range = None

        # update the plot data
        self.set_data(z, lut_range=lut_range)
        self.set_xy(x, y)

        # signal data changed and replot
        self.dataChanged.emit()

        if plot is not None:
            value = x[-1]
            axis = self.xAxis()
            xmin, xmax = plot.get_axis_limits(axis)
            if value > xmax or value < xmin:
                self.scrollRequested.emit(plot, axis, value)
            plot.update_colormap_axis(self)
            plot.replot()


class TaurusTrend2DScanItem(TaurusTrend2DItem):
    _xDataKey = 'point_nb'

    def __init__(self, channelKey, xDataKey, door, param=None, buffersize=512):
        TaurusTrend2DItem.__init__(
            self, param=param, buffersize=buffersize, stackMode=None)
        self._channelKey = channelKey
        self._xDataKey = xDataKey
        self.connectWithQDoor(door)

    def scanDataReceived(self, packet):
        '''
        packet is a dict with {type:str, "data":object} and the accepted types are: data_desc, record_data, record_end
        and the data objects are: seq<ColumnDesc.Todict()>, record.data dict and dict , respectively
        '''
        if packet is None:
            self.debug('Ignoring empty scan data packet')
            return
        id, packet = packet
        pcktype = packet.get("type", "__UNKNOWN_PCK_TYPE__")
        if pcktype == "data_desc":
            self._dataDescReceived(packet["data"])
        elif pcktype == "record_data":
            self._scanLineReceived(packet["data"])
        elif pcktype == "record_end":
            pass
        else:
            self.debug("Ignoring packet of type %s" % repr(pcktype))

    def clearTrend(self):
        self._yValues = None
        self._xBuffer = None
        self._zBuffer = None

    def _dataDescReceived(self, datadesc):
        '''prepares the plot according to the info in the datadesc dictionary'''
        self.clearTrend()
        # decide which data to use for x
        if self._xDataKey is None or self._xDataKey == "<mov>":
            self._autoXDataKey = datadesc['ref_moveables'][0]
        elif self._xDataKey == "<idx>":
            self._autoXDataKey = 'point_nb'
        else:
            self._autoXDataKey = self._xDataKey
        # set the x axis
        columndesc = datadesc.get('column_desc', [])
        xinfo = {'min_value': None, 'max_value': None}
        for e in columndesc:
            if e['label'] == self._autoXDataKey:
                xinfo = e
                break
        plot = self.plot()
        plot.set_axis_title('bottom', self._autoXDataKey)
        xmin, xmax = xinfo.get('min_value'), xinfo.get('max_value')
        if xmin is None or xmax is None:
            pass  # @todo: autoscale if any limit is unknown
        else:
            plot.set_axis_limits('bottom', xmin, xmax)

    def _scanLineReceived(self, recordData):
        '''Receives a recordData dictionary and updates the curves associated to it

        .. seealso:: <Sardana>/MacroServer/scan/scandata.py:Record.data

        '''
        # obtain the x value
        try:
            xval = recordData[self._autoXDataKey]
        except KeyError:
            self.warning(
                'Cannot find data "%s" in the current scan record. Ignoring', self._autoXDataKey)
            return
        if not numpy.isscalar(xval):
            self.warning('Data for "%s" is of type "%s". Cannot use it for the X values. Ignoring',
                         self._autoXDataKey, type(xval))
            return
        # obtain y value
        try:
            chval = recordData[self._channelKey]
        except KeyError:
            self.warning(
                'Cannot find data "%s" in the current scan record. Ignoring', self._channelKey)
        if chval.shape != self._yValues.shape:
            self.warning('Incompatible shape of "%s" (%s). Ignoring',
                         self._channelKey, repr(chval.shape))
            return

        # initialization
        if self._yValues is None:
            self._yValues = numpy.arange(chval.size, dtype='d')
        if self._xBuffer is None:
            self._xBuffer = ArrayBuffer(numpy.zeros(
                min(16, self.maxBufferSize), dtype='d'), maxSize=self.maxBufferSize)
        if self._zBuffer is None:
            self._zBuffer = ArrayBuffer(numpy.zeros(
                (min(16, self.maxBufferSize), chval.size), dtype='d'), maxSize=self.maxBufferSize)

        # update x
        self._xBuffer.append(xval)
        # update z
        self._zBuffer.append(chval)

        # check if there is enough data to start plotting
        if len(self._xBuffer) < 2:
            self.info('waiting for at least 2 values to start plotting')
            return

        x = self._xBuffer.contents()
        y = self._yValues
        z = self._zBuffer.contents().transpose()

        # update the plot data
        lut_range = self.get_lut_range()  # this is the range of the z axis (color scale)
        if lut_range[0] == lut_range[1]:
            # if the range was not set, make it None (autoscale z axis)
            lut_range = None
        self.set_data(z, lut_range=lut_range)
        self.set_xy(x, y)

        # signal data changed and replot
        self.dataChanged.emit()
        plot = self.plot()
        if plot is not None:
            value = x[-1]
            axis = self.xAxis()
            xmin, xmax = plot.get_axis_limits(axis)
            if value > xmax or value < xmin:
                self.scrollRequested.emit(plot, axis, value)
            plot.update_colormap_axis(self)
            plot.replot()

    def connectWithQDoor(self, doorname):
        '''connects this TaurusTrend2DScanItem to a QDoor

        :param doorname: (str) the QDoor name
        '''
        qdoor = taurus.Device(doorname)
        qdoor.recordDataUpdated.connect(self.scanDataReceived)

    def getModel(self):
        return self.__model

    def setModel(self, model):
        self.__model = model


def taurusImageMain():
    from guiqwt.tools import (RectangleTool, EllipseTool, HRangeTool, PlaceAxesTool,
                              MultiLineTool, FreeFormTool, SegmentTool, CircleTool,
                              AnnotatedRectangleTool, AnnotatedEllipseTool,
                              AnnotatedSegmentTool, AnnotatedCircleTool, LabelTool,
                              AnnotatedPointTool, ObliqueRectangleTool,
                              AnnotatedObliqueRectangleTool)
    try:  # In newer guiqwt versions, Annotated*CursorTool have been replaced by *CursorTool
        from guiqwt.tools import AnnotatedVCursorTool, AnnotatedHCursorTool
        VCursorTool, HCursorTool = AnnotatedVCursorTool, AnnotatedHCursorTool
    except ImportError:
        from guiqwt.tools import VCursorTool, HCursorTool

    from taurus.qt.qtgui.extra_guiqwt.tools import TaurusImageChooserTool
    from guiqwt.plot import ImageDialog
    from taurus.qt.qtgui.extra_guiqwt.builder import make
    from taurus.qt.qtgui.application import TaurusApplication
    import taurus.core.util.argparse
    import sys

    parser = taurus.core.util.argparse.get_taurus_parser()
    parser.set_usage("%prog [options] [<model1> [<model2>] ...]")
    parser.set_description("a taurus application for plotting 2D data sets")
    app = TaurusApplication(
        cmd_line_parser=parser, app_name="taurusimage", app_version=taurus.Release.version)
    args = app.get_command_line_args()

    # create a dialog with a plot and add the images
    win = ImageDialog(edit=False, toolbar=True, wintitle="Taurus Image",
                      options=dict(show_xsection=False, show_ysection=False))

    # add tools
    for toolklass in (TaurusImageChooserTool,
                      LabelTool, HRangeTool,
                      MultiLineTool, FreeFormTool, PlaceAxesTool,
                      AnnotatedObliqueRectangleTool,
                      AnnotatedEllipseTool, AnnotatedSegmentTool,
                      AnnotatedPointTool, VCursorTool,
                      HCursorTool):
        win.add_tool(toolklass)

    # add images from given models
    plot = win.get_plot()
    for m in args:
        img = make.image(taurusmodel=m)
        plot.add_item(img)
        # IMPORTANT: connect the cross section plots to the taurusimage so that
        # they are updated when the taurus data changes
        img.dataChanged.connect(win.update_cross_sections)

    win.exec_()


def test1():
    """Adapted from guiqwt cross_section.py example"""
    from guiqwt.plot import ImageDialog
    from taurus.qt.qtgui.extra_guiqwt.builder import make
    from taurus.qt.qtgui.application import TaurusApplication
    app = TaurusApplication(cmd_line_parser=None)

    # define a taurus image
    #model1 = 'sys/tg_test/1/short_image_ro'
    #model1 = 'sys/tg_test/1/long64_image_ro'
    model1 = 'sys/tg_test/1/ulong_image_ro'
    taurusimage = make.image(taurusmodel=model1)
    #taurusrgbimage = make.rgbimage(taurusmodel= 'eval:array([[[ 222, 0, 0], [0, 222, 0]], [[0, 0, 222], [222, 222, 222]]])')
    #taurusxyimage= make.xyimage(taurusmodel= model1)
    #taurusxyimage.set_xy(numpy.arange(251)*10,numpy.arange(251)*100 )

    # define normal image (guiqwt standard)
    #data = numpy.random.rand(100,100)
    #image = make.image(data=data)

    # create a dialog with a plot and add the images
    win = ImageDialog(edit=False, toolbar=True, wintitle="Taurus Cross sections test",
                      options=dict(show_xsection=False, show_ysection=False))
    from taurus.qt.qtgui.extra_guiqwt.tools import TaurusImageChooserTool
    win.add_tool(TaurusImageChooserTool)
    plot = win.get_plot()
    plot.add_item(taurusimage)
#    plot.add_item(taurusxyimage)
#    plot.add_item(image)
#    plot.add_item(taurusrgbimage)

#    win.get_itemlist_panel().show()

    # IMPORTANT: connect the cross section plots to the taurusimage so that they are updated when the taurus data changes
    # taurusimage.dataChanged.connect(win.update_cross_sections)

    win.exec_()

if __name__ == "__main__":
    test1()
    # taurusImageMain()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides an arrow based widget."""

from builtins import map
from builtins import range

import os
import math
import numpy

from taurus.external.qt import Qt
from taurus.core.units import Q_


__all__ = ["QWheelEdit"]

__docformat__ = 'restructuredtext'


class _ArrowButton(Qt.QPushButton):
    """Private class to be used by QWheelEdit for an arrow button"""

    ArrowPixmapName = "extra_icons:arrow01.svg"

    ButtonSize = 14
    IconSize = ButtonSize - 2

    def __init__(self, id, parent=None):
        Qt.QPushButton.__init__(self, parent)
        self.setAutoDefault(False)
        pixmap = self.getPixmap()
        self.setIcon(Qt.QIcon(pixmap))
        self.setIconSize(
            Qt.QSize(_ArrowButton.IconSize, _ArrowButton.IconSize))
        self.setFocusPolicy(Qt.Qt.ClickFocus)
        self.setFlat(True)
        self.setStyleSheet('_ArrowButton { border: 0px; }')
        self.setMaximumSize(_ArrowButton.ButtonSize, _ArrowButton.ButtonSize)
        #self.setMinimumSize(_ArrowButton.ButtonSize, _ArrowButton.ButtonSize)
        self._id = id
        self._inc = math.pow(10, id)


class _UpArrowButton(_ArrowButton):
    """Private class to be used by QWheelEdit for an up arrow button"""

    ArrowPixmapKey = "upArrow01"

    def __init__(self, id, parent=None):
        _ArrowButton.__init__(self, id, parent)

    def getPixmap(self):
        pm = Qt.QPixmapCache.find(_UpArrowButton.ArrowPixmapKey)
        if pm is None:
            pm = Qt.QPixmap(self.ArrowPixmapName)
            Qt.QPixmapCache.insert(_UpArrowButton.ArrowPixmapKey, pm)
        return pm


class _DownArrowButton(_ArrowButton):
    """Private class to be used by QWheelEdit for a down arrow button"""

    ArrowPixmapKey = "downArrow01"

    def __init__(self, id, parent=None):
        _ArrowButton.__init__(self, id, parent)
        self._inc = -self._inc

    def getPixmap(self):
        pm = Qt.QPixmapCache.find(_DownArrowButton.ArrowPixmapKey)
        if pm is None:
            pm = Qt.QPixmap(self.ArrowPixmapName)
            pm = pm.transformed(Qt.QTransform().rotate(180))
            Qt.QPixmapCache.insert(_DownArrowButton.ArrowPixmapKey, pm)
        return pm

class _DigitLabel(Qt.QLabel):
    """A private single digit label to be used by QWheelEdit widget"""

    PixmapKeys = list(map(str, range(10))) + ['blank', 'minus', 'point']

    def __init__(self, lbl, parent=None):
        Qt.QLabel.__init__(self, parent)
        self.setAlignment(Qt.Qt.AlignCenter)
        self.setFocusPolicy(Qt.Qt.StrongFocus)
        self.setStyleSheet(
            """QLabel:focus {background-color: rgb(180,180,255); border-width: 1px; border-style: solid; border-color: rgb(200,200,255);}""")
        self.resetSkin()
        self.setText(lbl)
        self._upButton = None
        self._downButton = None

    def setButtons(self, up, down):
        self._upButton = up
        self._downButton = down

    def keyPressEvent(self, key_event):
        if key_event.key() == Qt.Qt.Key_Up:
            self._upButton.click()
        elif key_event.key() == Qt.Qt.Key_Down:
            self._downButton.click()
        elif key_event.key() == Qt.Qt.Key_Right:
            self.focusNextChild()
        elif key_event.key() == Qt.Qt.Key_Left:
            self.focusPreviousChild()
        Qt.QLabel.keyPressEvent(self, key_event)

    def _update_skin_cache(self):
        skin = os.path.join('digits', self.getSkin())

        for i, d in enumerate(self.PixmapKeys):
            full_name = os.path.join(skin, d)
            pm = Qt.QPixmapCache.find(full_name)
            if pm is None:
                file_name = '%s.png' % full_name
                pm = Qt.QPixmap(file_name)
                if pm is None:
                    continue
                if not pm.isNull():
                    Qt.QPixmapCache.insert(full_name, pm)

    def setSkin(self, skin_name):
        self._skin = skin_name
        self._update_skin_cache()
        self.update()

    def getSkin(self):
        return self._skin

    def resetSkin(self):
        self.setSkin('Nixie')

    def setText(self, t):
        k = t
        if k == '.':
            k = 'point'
        elif k == '-':
            k = 'minus'
        elif k is None or k == '+':
            k = 'blank'

        k = os.path.join('digits', self.getSkin(), k)
        pm = Qt.QPixmapCache.find(k)
        if pm is None:
            Qt.QLabel.setText(self, t)
            return
        Qt.QLabel.setText(self, '')
        self.setPixmap(pm)

    skin = Qt.pyqtProperty('QString', getSkin, setSkin, resetSkin)


class QWheelEdit(Qt.QFrame):
    """A widget designed to handle numeric scalar values. It allows interaction
    based on single digit as well as normal value edition."""

    numberChanged = Qt.pyqtSignal(float)
    numberEdited = Qt.pyqtSignal(float)
    returnPressed = Qt.pyqtSignal()

    DefaultIntDigitCount = 6
    DefaultDecDigitCount = 2

    def __init__(self, parent=None):
        """__init__(self, parent = None) -> QWheelEdit

        Constructor

        @param[in] parent (QWidget) the parent widget (optional, default is None
                          meaning there is no parent
        """
        Qt.QFrame.__init__(self, parent)
        Qt.QGridLayout(self)
        self._roundFunc = None
        self._previous_value = 0
        self._value = 0
        self._value_str = '0'
        self._minValue = numpy.finfo('d').min  # -inf
        self._maxValue = numpy.finfo('d').max  # inf
        self._editor = None
        self._editing = False
        self._hideEditWidget = True
        self._showArrowButtons = True
        self._forwardReturn = False
        self._validator = Qt.QDoubleValidator(self)
        self._validator.setNotation(self._validator.StandardNotation)
        self._setDigits(QWheelEdit.DefaultIntDigitCount,
                        QWheelEdit.DefaultDecDigitCount)
        self._setValue(0)

        self._build()
        self.setDigitCountAction = Qt.QAction("Change digits", self)
        self.setDigitCountAction.triggered.connect(self._onSetDigitCount)
        self.addAction(self.setDigitCountAction)
        self.setContextMenuPolicy(Qt.Qt.ActionsContextMenu)

    def _getMinPossibleValue(self):
        """_getMinPossibleValue(self) -> None

        Determines which is the minimum possible value that can be represented
        with the current total number of digits.

        @return (float) the minimum possible value
        """
        decmax = 0
        for i in range(self.getDecDigitCount()):
            decmax += 9 * math.pow(10, -(i + 1))
        return -math.pow(10.0, self.getIntDigitCount()) + 1 - decmax

    def _getMaxPossibleValue(self):
        """_getMaxPossibleValue(self) -> None

        Determines which is the maximum possible value that can be represented
        with the current total number of digits.

        @return (float) the maximum possible value
        """
        decmax = 0
        for i in range(self.getDecDigitCount()):
            decmax += 9 * math.pow(10, -(i + 1))
        return math.pow(10.0, self.getIntDigitCount()) - 1 + decmax

    def _build(self):
        """_build(self) -> None

        Builds this widget sub-items"""

        l = self.layout()
        l.setSpacing(0)
        l.setContentsMargins(0, 0, 0, 0)

        id = self.getIntDigitCount()
        dd = self.getDecDigitCount()
        digits = self.getDigitCount()

        self._upButtons = Qt.QButtonGroup()
        self._downButtons = Qt.QButtonGroup()
        self._digitLabels = []
        for l in self._digitLabels:
            l.setAlignment(Qt.AlignCenter)

        showDot = self.getDecDigitCount() > 0

        signLabel = _DigitLabel('+')
        signLabel.setFocusPolicy(Qt.Qt.NoFocus)
        signLabel.setAlignment(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter)
        self._digitLabels.append(signLabel)
        l.addWidget(signLabel, 1, 0)
        l.setRowMinimumHeight(1, signLabel.minimumSizeHint().height())
        l.setColumnMinimumWidth(0, _ArrowButton.ButtonSize)
        l.setColumnStretch(0, 1)

        for i in range(id):
            col = i + 1
            d = _DigitLabel('0')
            up = _UpArrowButton(id - i - 1)
            down = _DownArrowButton(id - i - 1)
            d.setButtons(up, down)
            up.setFocusProxy(d)
            down.setFocusProxy(d)
            self._upButtons.addButton(up, col)
            self._downButtons.addButton(down, col)
            self._digitLabels.append(d)
            if self.getShowArrowButtons():
                l.addWidget(up, 0, col)
                l.addWidget(down, 2, col)
            l.addWidget(d, 1, col)

        if showDot:
            dotLabel = _DigitLabel('.')
            dotLabel.setFocusPolicy(Qt.Qt.NoFocus)
            dotLabel.setAlignment(Qt.Qt.AlignCenter)
            self._digitLabels.append(dotLabel)
            l.addWidget(dotLabel, 1, id + 1)

        for i in range(id, digits):
            col = i + 1
            if showDot:
                col += 1
            d = _DigitLabel('0')
            up = _UpArrowButton(id - i - 1)
            down = _DownArrowButton(id - i - 1)
            d.setButtons(up, down)
            up.setFocusProxy(d)
            down.setFocusProxy(d)
            self._upButtons.addButton(up, col)
            self._downButtons.addButton(down, col)
            self._digitLabels.append(d)
            if self.getShowArrowButtons():
                l.addWidget(up, 0, col)
                l.addWidget(down, 2, col)
            l.addWidget(d, 1, col)

        self._upButtons.buttonClicked.connect(self.buttonPressed)
        self._downButtons.buttonClicked.connect(self.buttonPressed)

        ed = Qt.QLineEdit(self)
        ed.setFrame(False)
        ed.returnPressed.connect(self.editingFinished)
        ed.editingFinished.connect(self.hideEditWidget)
        rect = Qt.QRect(l.cellRect(1, 0).topLeft(),
                        l.cellRect(1, l.columnCount() - 1).bottomRight())
        ed.setGeometry(rect)
        ed.setAlignment(Qt.Qt.AlignRight)
        ed.setValidator(self._validator)
        ed.setVisible(False)
        self._editor = ed

        self._updateValidator()

        # set the minimum height for the widget
        # (otherwise the hints seem to be ignored by the layouts)
        min_height = max(ed.minimumSizeHint().height(),
                         signLabel.minimumSizeHint().height())
        if self.getShowArrowButtons():
            min_height += 2 * _ArrowButton.ButtonSize
        self.setMinimumHeight(min_height)

        self.clearWarning()

    def _clear(self):
        """_clear(self) -> None

        Clears this widget sub-items"""

        self._upButtons.buttonClicked.disconnect(self.buttonPressed)
        self._downButtons.buttonClicked.disconnect(self.buttonPressed)

        for b in self._upButtons.buttons():
            self._upButtons.removeButton(b)
            b.setParent(None)
            b.destroy()
        for b in self._downButtons.buttons():
            self._downButtons.removeButton(b)
            b.setParent(None)
            b.destroy()
        for label in self._digitLabels:
            label.setParent(None)
            label.destroy()
        self._digitLabels = None
        self._upButtons.setParent(None)
        self._downButtons.setParent(None)
        self._upButtons = None
        self._downButtons = None
        self._editor.setParent(None)
        self._editor.destroy()
        self._editor = None

    def _setDigits(self, int_nb=None, dec_nb=None):
        """_setDigits(self, int_nb=None, dec_nb=None) -> None

        Sets the number of digits that this widget shows. It will ensure that
        the current value can be displayed (i.e., int_nb will be forced to be
        large enough for displaying the integer representation of the value)

        @param[in] int_nb (int) number of integer digits (optional, default is
                   None, meaning use the existing number of integer digits
        @param[in] dec_nb (int) number of decimal digits (optional, default is
                   None, meaning use the existing number of decimal digits
        """

        if not int_nb is None:
            self._intDigitCount = int_nb
        if not dec_nb is None:
            self._decDigitCount = dec_nb

        # make sure that the current value can be displayed
        self._intDigitCount = max(self._intDigitCount, len("%d" % self._value))

        self._digitCount = self._intDigitCount + self._decDigitCount
        total_chars = self._digitCount
        total_chars += 1  # for sign
        if self._decDigitCount > 0:
            total_chars += 1  # for dot
        self._valueFormat = '%%+0%d.%df' % (total_chars, self._decDigitCount)
        self._updateValidator()
        # we call setValue to update the self._value_str
        self._setValue(self.getValue())

    def _buildValueStr(self, v):
        """_buildValueStr(self, v) -> str

        Builds a string representation of the given value according to the
        number of digits this widget can currently represent

        @param[in] v (float) the value to be translated

        @return (str) a proper string representation of the given value
        """
        if v is None:
            ret = (self._valueFormat % 0).replace('0', '-')
        else:
            ret = self._valueFormat % v
            if ret.endswith('nan'):
                ret = ret.replace('0', ' ')
        self._value_str = ret
        return ret

    def _updateDigits(self):
        """_updateDigits(self) -> None

        Updates this widget contents (sub-widgets) according to the current
        widget configuration
        """
        self._clear()
        self._build()
        self._updateValue(False, False)

    def _updateValue(self, trigValueChanged=True, trigValueEdited=True):
        """_updateValue(self, trigValueChanged=True, trigValueEdited=True) -> None

        Updates this widget displayed value and potentially send signal(s)

        @param[in] trigValueChanged (bool) (optional, default is True) wheather or
                   not to send signal 'valueChanged(double)'
        @param[in] trigValueEdited (bool) (optional, default is True) wheather or
                   not to send signal 'valueEdit(double)'
        """
        v, v_str = self.getValue(), self.getValueStr()

        if len(v_str) > len(self._digitLabels):
            # do auto adjust
            if '.' in v_str:
                dc = list(map(len, v_str.split('.')))
            else:
                dc = len(v_str), 0
            self._setDigits(*dc)
            v_str = self._buildValueStr(v)
            self._clear()
            self._build()

        for i, c in enumerate(v_str):
            self._digitLabels[i].setText(c)

        if trigValueChanged:
            self.numberChanged.emit(v)

        if trigValueEdited:
            self.numberEdited.emit(v)

    def setRoundFunc(self, roundFunc):
        """setRoundFunc(self, roundFunc) -> None

        Sets the rounding function to use when calling _setValue(). This allows you to
        filter invalid user input

        @param[in] roundFunc (callable) the rounding function to use
        """
        self._roundFunc = roundFunc

    def getPreviousValue(self):
        """getPreviousValue(self) -> float

        Gives the previous value of this widget

        @return (float) the previous value of this widget
        """
        return self._previous_value

    def _setValue(self, v):
        """_setValue(self, v) -> None

        Sets value of this widget. If the given value exceeds any limit, the
        value is NOT set.
        """
        if v is None:
            return
        if self._roundFunc:
            v = self._roundFunc(v)
        if v > self._validator.top() or v < self._validator.bottom():
            Qt.QMessageBox.warning(
                self,
                "Invalid Value",
                "'{}' is out of the allowed range ({}, {})".format(
                    v,
                    self._validator.bottom(),
                    self._validator.top(),
                )
            )
            return
        self._previous_value = self._value
        self._value = v

        str_value = self._buildValueStr(v)
        ed = self.getEditWidget()
        if ed is not None:
            ed.setText(str_value)


    def setWarning(self, msg):
        """setWarning(self, msg) -> None

        Activates the warning style for this widget. This means a violet border
        and a tooltip with the given message.

        @param[in] msg (str) the message to be displayed as tooltip
        """
        self.setStyleSheet(
            'QWheelEdit {border: 1px solid; border-radius: 4px; border-color: violet}')
        self.setToolTip(msg)

    def clearWarning(self):
        """clearWarning(self) -> None

        Clears the warning style. If not in warning mode, nothing is done.
        """
        self.setStyleSheet(
            'QWheelEdit {border: 1px solid; border-radius: 4px; border-color: rgba(0,0,0,0)}')
        self.setToolTip('')

    def buttonPressed(self, b):
        """buttonPressed(self, b) -> None

        Slot executed when an arrow button is pressed from the button group

        @param[in] b (_ArrowButton) the button which was pressed
        """
        if self._editing:
            self.hideEditWidget()
        else:
            self._setValue(self.getValue() + b._inc)
            self._updateValue()

    def _onSetDigitCount(self):
        text, ok = Qt.QInputDialog.getText(
            self,
            "Change digits",
            "Enter digits as <int_nb>.<dec_nb>",
            text="{:d}.{:d}".format(
                self.getIntDigitCount(),
                self.getDecDigitCount()
            )
        )
        if ok:
            try:
                dc = list(map(int, text.split('.')))
                self.setDigitCount(*dc)
            except Exception as e:
                Qt.QMessageBox.warning(
                    self,
                    "Invalid digit count specification",
                    'Invalid specification: "{}"\nReason:{}'.format(text, e)
                )
                self._onSetDigitCount()

    def setDigitCount(self, int_nb, dec_nb):
        """setDigitCount(self, int_nb, dec_nb) -> None

        Updates the displayed digits.

        @param[in] int_nb(int) number of integer digits
        @param[in] dec_nb(int) number of decimal digits
        """
        self._setDigits(int_nb=int_nb, dec_nb=dec_nb)
        self._updateDigits()

    def getDigitCount(self):
        """getDigitCount(self) -> int

        Gets the total number of digits this widget displays

        @return (int) the total number of digits this widget displays
        """
        return self._digitCount

    def getIntDigitCount(self):
        """getIntDigitCount(self) -> int

        Gets the number of integer digits this widget displays

        @return (int) the number of integer digits this widget displays
        """
        return self._intDigitCount

    def setIntDigitCount(self, n):
        """setIntDigitCount(self, n) -> None

        Sets the number of integer digits this widget displays

        @param[in] n (int) the number of integer digits to display
        """
        self._setDigits(int_nb=n)
        self._updateDigits()

    def resetIntDigitCount(self):
        """resetIntDigitCount(self) -> None

        Resets the number of integer digits this widget displays to DefaultIntDigitCount
        """
        self.setIntDigitCount(QWheelEdit.DefaultIntDigitCount)

    def getDecDigitCount(self):
        """getDecDigitCount(self) -> int

        Gets the number of decimal digits this widget displays

        @return (int) the number of decimal digits this widget displays
        """
        return self._decDigitCount

    def setDecDigitCount(self, n):
        """setDecDigitCount(self, n) -> None

        Sets the number of decimal digits this widget displays

        @param[in] n (int) the number of decimal digits to display
        """

        self._setDigits(dec_nb=n)
        self._updateDigits()

    def resetDecDigitCount(self):
        """resetDecDigitCount(self) -> None

        Resets the number of decimal digits this widget displays to DefaultDecDigitCount
        """
        self.setDecDigitCount(QWheelEdit.DefaultDecDigitCount)

    def setValue(self, v):
        """setValue(self, v) -> None

        Sets the value of this widget.
        Send a 'valueChanged(double)' Qt signal

        @param[in] v (float/Quantity) the value to be set
        """
        if isinstance(v, Q_):
            v = v.magnitude
        self._setValue(v)
        self._updateValue(trigValueEdited=False)

    def resetValue(self):
        """resetValue(self) -> None

        Resets the value of this widget to 0.0
        """
        self.setValue(0.0)

    def getValue(self):
        """getValue(self) -> float

        Gets the current value of this widget

        @return (float) the value currently displayed by the widget
        """
        return self._value

    def getValueStr(self):
        """getValueStr(self) -> str

        Gets the current value string of this widget

        @return (str) the value currently displayed by the widget
        """
        return self._value_str

    def getMinValue(self):
        """getMinValue(self) -> float

        Gets the minimum allowed value

        @return (float) the minimum allowed value
        """
        return self._minValue

    def setMinValue(self, v):
        """setMinValue(self, v) -> None

        Sets the minimum allowed value for the widget

        @param[in] v (float) the new minimum allowed value
        """
        self._minValue = v
        self._updateValidator()

    def _updateValidator(self):
        min_ = max(self._minValue, self._getMinPossibleValue())
        max_ = min(self._maxValue, self._getMaxPossibleValue())
        self._validator.setRange(min_, max_, self._decDigitCount)

    def resetMinValue(self):
        """resetMinValue(self) -> None

        Resets the minimum allowed value to -inf
        """
        self.setMinValue(numpy.finfo('d').min)

    def getMaxValue(self):
        """getMaxValue(self) -> float

        Gets the maximum allowed value

        @return (float) the maximum allowed value
        """
        return self._maxValue

    def setMaxValue(self, v):
        """setMaxValue(self, v) -> None

        Sets the maximum allowed value for the widget

        @param[in] v (float) the new maximum allowed value
        """
        self._maxValue = v
        self._updateValidator()

    def resetMaxValue(self):
        """resetMaxValue(self) -> None

        Resets the maximum allowed value to +inf
        """
        self.setMaxValue(numpy.finfo('d').max)

    def getAutoRepeat(self):
        return self._upButtons.buttons()[0].autoRepeat()

    def setAutoRepeat(self, v):
        for b in self._upButtons.buttons():
            b.setAutoRepeat(v)
        for b in self._downButtons.buttons():
            b.setAutoRepeat(v)

    def resetAutoRepeat(self):
        self.setAutoRepeat(False)

    def getAutoRepeatDelay(self):
        return self._upButtons.buttons()[0].autoRepeatDelay()

    def setAutoRepeatDelay(self, milisecs):
        for b in self._upButtons.buttons():
            b.setAutoRepeatDelay(milisecs)
        for b in self._downButtons.buttons():
            b.setAutoRepeatDelay(milisecs)

    def getAutoRepeatInterval(self):
        return self._upButtons.buttons()[0].autoRepeatInterval()

    def setAutoRepeatInterval(self, milisecs):
        for b in self._upButtons.buttons():
            b.setAutoRepeatInterval(milisecs)
        for b in self._downButtons.buttons():
            b.setAutoRepeatInterval(milisecs)

    def getEditWidget(self):
        """getEditWidget(self) -> QWidget

        Gets the widget object used when the user manually sets the value

        @return (QWidget) the widget used for editing
        """
        return self._editor

    def editingFinished(self):
        """editingFinished(self) -> None

        Slot called when the user finishes editing
        """
        ed = self.getEditWidget()
        v = float(ed.text())
        self._setValue(v)
        self._updateValue()

    def showEditWidget(self):
        """showEditWidget(self) -> None

        Forces the edition widget to be displayed
        """
        ed, l = self.getEditWidget(), self.layout()
        rect = Qt.QRect(l.cellRect(1, 0).topLeft(), l.cellRect(1,
                                                               l.columnCount() - 1).bottomRight())
        ed.setGeometry(rect)
        ed.setAlignment(Qt.Qt.AlignRight)
        ed.setText(self.getValueStr())
        ed.selectAll()
        ed.setFocus()
        ed.setVisible(True)
        self._editing = True

    def hideEditWidget(self):
        """hideEditWidget(self) -> None

        Forces the edition widget to be hidden
        """
        ed = self.getEditWidget()
        ed.setVisible(False)
        self._editing = False
        self.setFocus()

    def wheelEvent(self, evt):
        numDegrees = evt.delta() // 8
        numSteps = numDegrees // 15
        #w = Qt.QApplication.focusWidget()
        w = self.focusWidget()
        if not isinstance(w, _DigitLabel):
            return Qt.QFrame.wheelEvent(self, evt)

        if numSteps > 0:
            inc = w._upButton._inc
        else:
            inc = -w._downButton._inc
        numSteps *= inc
        self.setValue(self.getValue() + numSteps)
        evt.accept()

    def mouseDoubleClickEvent(self, mouse_event):
        """mouseDoubleClickEvent(self, mouse_event)

        Executed when user presses double click. This widget shows the edition
        widget when this happens
        """
        self.showEditWidget()

    def keyPressEvent(self, key_event):
        """keyPressEvent(self, key_event) -> None

        Exectuted when the user presses a key.
        F2 enters/leaves edition mode. ESC leaves edition mode
        """
        k = key_event.key()
        if k == Qt.Qt.Key_F2:
            if self._editing:
                self.hideEditWidget()
            else:
                self.showEditWidget()
            return
        elif k == Qt.Qt.Key_Escape:
            if self._editing:
                self.hideEditWidget()
                return
        elif k in (Qt.Qt.Key_Return, Qt.Qt.Key_Enter):
            if self._editing:
                self.hideEditWidget()
            else:
                self.returnPressed.emit()

        # TODO Decide when to emit editingFinished for completeness
        Qt.QWidget.keyPressEvent(self, key_event)

    def getShowArrowButtons(self):
        return self._showArrowButtons

    def setShowArrowButtons(self, yesno):
        self._showArrowButtons = yesno
        self._updateDigits()

    def resetShowArrowButtons(self):
        self.setShowArrowButtons(True)

    def getHideEditWidget(self):
        """getHideEditWidget(self) -> bool

        Gets the info if edition widget should be hidden when 'focusOut' event
        occurs.

        @return (bool)
        """
        return self._hideEditWidget

    def setHideEditWidget(self, focus_out=True):
        """setFocusOut(self, focus_out=True) -> None

        Sets if edition widget should be hidden when 'focusOut' event occurs.
        If set to False, edition widget is hidden only when 'F2', 'Esc',
        'Enter' and arrow button are pressed. Default set to True.

        @param[in] focus_out (bool) whether or not to hide edition widget
        after 'focusOut' event.
        """
        if focus_out and not self._hideEditWidget:
            ed = self.getEditWidget()
            ed.editingFinished.connect(self.hideEditWidget)
            self._hideEditWidget = True
        elif not focus_out and self._hideEditWidget:
            ed = self.getEditWidget()
            ed.editingFinished.disconnect(self.hideEditWidget)
            self._hideEditWidget = False

    def isReturnForwarded(self):
        """isReturnForwarded(self) -> bool

        Gets the info if returnPressed is forwarded.

        @return (bool)
        """
        return self._forwardReturn

    def setReturnForwarded(self, forward_rtn=False):
        """setReturnForwarded(self, forward_rtn=False) -> None

        Sets forwarding of returnPressed. If set to True, returnPressed from
        edition widget emits returnPressed of 'QWheelEdit' widget.

        @param[in] forward_rtn (bool) whether or not to forward returnPressed
        signal
        """
        if forward_rtn and not self._forwardReturn:
            ed = self.getEditWidget()
            ed.returnPressed.connect(self.returnPressed)
            self._forwardReturn = True

        elif not forward_rtn and self._forwardReturn:
            ed = self.getEditWidget()
            ed.returnPressed.disconnect(self.returnPressed)
            self._forwardReturn = False

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT properties
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    showArrowButtons = Qt.pyqtProperty("bool", getShowArrowButtons,
                                       setShowArrowButtons,
                                       resetShowArrowButtons)

    integerDigits = Qt.pyqtProperty("int", getIntDigitCount,
                                    setIntDigitCount, resetIntDigitCount)

    decimalDigits = Qt.pyqtProperty("int", getDecDigitCount,
                                    setDecDigitCount, resetDecDigitCount)

    value = Qt.pyqtProperty("double", getValue,
                            setValue, resetValue)

    minValue = Qt.pyqtProperty("double", getMinValue,
                               setMinValue, resetMinValue)

    maxValue = Qt.pyqtProperty("double", getMaxValue,
                               setMaxValue, resetMaxValue)

    autoRepeat = Qt.pyqtProperty("bool", getAutoRepeat,
                                 setAutoRepeat, resetAutoRepeat)

    autoRepeatDelay = Qt.pyqtProperty("int", getAutoRepeatDelay,
                                      setAutoRepeatDelay)

    autoRepeatInterval = Qt.pyqtProperty("int", getAutoRepeatInterval,
                                         setAutoRepeatInterval)


def main():
    import taurus.qt.qtgui.icon  # otherwise the arrows don't show in the demo
    global arrowWidget

    def resetAll():
        arrowWidget.resetIntDigitCount()
        arrowWidget.resetDecDigitCount()
        arrowWidget.resetMinValue()
        arrowWidget.resetMaxValue()
        arrowWidget.resetValue()

    def setNAN():
        arrowWidget.setValue(float('nan'))

    def setNone():
        arrowWidget.setValue(None)

    a = Qt.QApplication([])
    panel = Qt.QWidget()
    l = Qt.QFormLayout(panel)
    button_layout = Qt.QVBoxLayout()
    arrowWidget = QWheelEdit(panel)
    isb, dsb = Qt.QSpinBox(panel), Qt.QSpinBox(panel)
    minv, maxv = Qt.QDoubleSpinBox(panel), Qt.QDoubleSpinBox(panel)
    resetbutton = Qt.QPushButton("Reset", panel)
    resetbutton.setDefault(True)
    nanbutton = Qt.QPushButton("Set NAN", panel)
    nonebutton = Qt.QPushButton("Set None", panel)
    hideEditCB = Qt.QCheckBox()
    hideEditCB.setChecked(True)
    showarrowbutton = Qt.QCheckBox("", panel)

    l.addRow("Value", arrowWidget)
    l.addRow("Integer digits:", isb)
    l.addRow("Decimal digits:", dsb)
    l.addRow("Minimum value:", minv)
    l.addRow("Maximum value:", maxv)
    l.addRow("Show arrows:", showarrowbutton)
    l.addRow("hideEditCB", hideEditCB)
    l.addRow(button_layout)
    button_layout.addWidget(nanbutton)
    button_layout.addWidget(nonebutton)
    button_layout.addWidget(resetbutton)
    isb.setValue(arrowWidget.getIntDigitCount())
    dsb.setValue(arrowWidget.getDecDigitCount())
    minv.setRange(numpy.finfo('d').min, numpy.finfo('d').max)
    maxv.setRange(numpy.finfo('d').min, numpy.finfo('d').max)
    minv.setValue(arrowWidget.getMinValue())
    maxv.setValue(arrowWidget.getMaxValue())
    showarrowbutton.setChecked(arrowWidget.getShowArrowButtons())
    isb.valueChanged.connect(arrowWidget.setIntDigitCount)
    dsb.valueChanged.connect(arrowWidget.setDecDigitCount)
    minv.valueChanged.connect(arrowWidget.setMinValue)
    showarrowbutton.stateChanged.connect(arrowWidget.setShowArrowButtons)
    nanbutton.clicked.connect(setNAN)
    nonebutton.clicked.connect(setNone)
    hideEditCB.toggled.connect(arrowWidget.setHideEditWidget)
    resetbutton.clicked.connect(resetAll)
    panel.setVisible(True)
    a.exec_()

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
arrayedit.py: Widget for editing a spectrum/array via control points
"""
from __future__ import absolute_import

from builtins import range
import numpy
from taurus.external.qt import Qt, Qwt5
from taurus.qt.qtgui.util.ui import UILoadable
from .curvesAppearanceChooserDlg import CurveAppearanceProperties


@UILoadable
class ControllerBox(Qt.QWidget):

    selected = Qt.pyqtSignal(int)

    def __init__(self, parent=None, x=0, y=0, corr=0):
        Qt.QWidget.__init__(self, parent)
        self.loadUi()
        self._x = x
        self.setY(y)
        self.box.setTitle('x=%6g' % self._x)
        self.corrSB.setValue(corr)
        self.ctrlObj = self.corrSB.ctrlObj = self.lCopyBT.ctrlObj = self.rCopyBT.ctrlObj = self.lScaleBT.ctrlObj = self.rScaleBT.ctrlObj = self
        # reimplementing the focusInEvent method for the spinbox
        self.corrSB.focusInEvent = self.corrSB_focusInEvent
        self.box.mousePressEvent = self.mousePressEvent
        #self.connect(self.corrSB, Qt.SIGNAL('valueChanged(double)'), self.enableScale)

    def mousePressEvent(self, event):
        self.selected.emit(self._x)
        # print 'SELECTED', self
        #Qt.QDoubleSpinBox.focusInEvent(self.corrSB, event)

    def corrSB_focusInEvent(self, event):
        self.selected.emit(self._x)
        # print 'GOT FOCUS', self
        Qt.QDoubleSpinBox.focusInEvent(self.corrSB, event)

    def setY(self, y):
        self._y = y
        self.enableScale()

    def enableScale(self, *args):
        enable = (self._y + self.corrSB.value()) != 0
        self.lScaleBT.setEnabled(enable)
        self.rScaleBT.setEnabled(enable)


@UILoadable
class EditCPointsDialog(Qt.QDialog):

    def __init__(self, parent=None, x=0):
        Qt.QDialog.__init__(self, parent)
        self.loadUi()


@UILoadable
class AddCPointsDialog(Qt.QDialog):

    def __init__(self, parent=None, x=0):
        Qt.QDialog.__init__(self, parent)
        self.loadUi()


@UILoadable
class ArrayEditor(Qt.QWidget):

    def __init__(self, parent=None):
        Qt.QWidget.__init__(self, parent)
        self.loadUi()

        self._controllers = []

        # construct the layout for controllers container
        self.ctrlLayout = Qt.QHBoxLayout(self.controllersContainer)
        self.ctrlLayout.setContentsMargins(5, 0, 5, 0)
        self.ctrlLayout.setSpacing(1)

        # implement scroll bars for the controllers container
        self.scrollArea = Qt.QScrollArea(self)
        self.scrollArea.setWidget(self.controllersContainer)
        self.scrollArea.setVerticalScrollBarPolicy(Qt.Qt.ScrollBarAlwaysOff)
        self.scrollArea.setWidgetResizable(True)
        self.cpointsGroupBox.layout().insertWidget(0, self.scrollArea)

        # initialize data
        cpoints = 2
        self.x = numpy.arange(256, dtype='double')
        self.y = numpy.zeros(256, dtype='double')
        self.xp = numpy.linspace(self.x[0], self.x[-1], cpoints)
        self.corrp = numpy.zeros(cpoints)
        self.yp = numpy.interp(self.xp, self.x, self.y)
        self.corr = numpy.zeros(self.x.size)

        # markers
        self.markerPos = self.xp[0]
        self.marker1 = Qwt5.QwtPlotMarker()
        self.marker1.setSymbol(Qwt5.QwtSymbol(Qwt5.QwtSymbol.Rect,
                                              Qt.QBrush(Qt.Qt.NoBrush),
                                              Qt.QPen(Qt.Qt.green),
                                              Qt.QSize(8, 8)))
        self.marker1.attach(self.plot1)
        self.marker2 = Qwt5.QwtPlotMarker()
        self.marker2.setSymbol(Qwt5.QwtSymbol(Qwt5.QwtSymbol.Rect,
                                              Qt.QBrush(Qt.Qt.NoBrush),
                                              Qt.QPen(Qt.Qt.green),
                                              Qt.QSize(8, 8)))
        self.marker2.attach(self.plot2)

        # cpointsPickers
        self._cpointMovingIndex = None
        self._cpointsPicker1 = Qwt5.QwtPicker(self.plot1.canvas())
        self._cpointsPicker1.setSelectionFlags(Qwt5.QwtPicker.PointSelection)
        self._cpointsPicker2 = Qwt5.QwtPicker(self.plot2.canvas())
        self._cpointsPicker2.setSelectionFlags(Qwt5.QwtPicker.PointSelection)
        self._cpointsPicker1.widgetMousePressEvent = self.plot1MousePressEvent
        self._cpointsPicker1.widgetMouseReleaseEvent = self.plot1MouseReleaseEvent
        self._cpointsPicker2.widgetMousePressEvent = self.plot2MousePressEvent
        self._cpointsPicker2.widgetMouseReleaseEvent = self.plot2MouseReleaseEvent
        self._cpointsPicker1.widgetMouseDoubleClickEvent = self.plot1MouseDoubleClickEvent
        self._cpointsPicker2.widgetMouseDoubleClickEvent = self.plot2MouseDoubleClickEvent
        self._populatePlots()
        self.resetCorrection()
        self._selectedController = self._controllers[0]

        self._addCPointsDialog = AddCPointsDialog(self)

        # Launch low-priority initializations (to speed up load time)
        # Qt.QTimer.singleShot(0, <method>)

        # connections
        self.addCPointsBT.clicked.connect(self._addCPointsDialog.show)
        self._addCPointsDialog.editBT.clicked.connect(self.showEditCPointsDialog)
        self._addCPointsDialog.cleanBT.clicked.connect(self.resetCorrection)
        self._addCPointsDialog.addSingleCPointBT.clicked.connect(self.onAddSingleCPointBT)
        self._addCPointsDialog.addRegEspCPointsBT.clicked.connect(self.onAddRegEspCPointsBT)

    def plot1MousePressEvent(self, event):
        self.plotMousePressEvent(event, self.plot1)

    def plot2MousePressEvent(self, event):
        self.plotMousePressEvent(event, self.plot2)

    def plotMousePressEvent(self, event, taurusplot):
        picked, pickedCurveName, pickedIndex = taurusplot.pickDataPoint(
            event.pos(), scope=20, showMarker=False, targetCurveNames=['Control Points'])
        if picked is not None:
            self.changeCPointSelection(picked.x())
            self.makeControllerVisible(self._controllers[pickedIndex])
            self._cpointMovingIndex = pickedIndex
            self._movingStartYPos = event.y()
            taurusplot.canvas().setCursor(Qt.Qt.SizeVerCursor)

    def plot1MouseReleaseEvent(self, event):
        self.plotMouseReleaseEvent(event, self.plot1)

    def plot2MouseReleaseEvent(self, event):
        self.plotMouseReleaseEvent(event, self.plot2)

    def plotMouseReleaseEvent(self, event, taurusplot):
        if self._cpointMovingIndex is None:
            return  # if no cpoint was picked, do nothing on release
        # no motion s performed if the y position is unchanged or if the mouse
        # release is out of the canvas
        validMotion = (self._movingStartYPos != event.pos().y()
                       ) and taurusplot.canvas().rect().contains(event.pos())
        if validMotion:
            # calculate the new correction
            newCorr = taurusplot.invTransform(
                taurusplot.getCurve('Control Points').yAxis(), event.y())
            if taurusplot is self.plot1:
                newCorr -= self.yp[self._cpointMovingIndex]
            # apply new correction
            self._controllers[self._cpointMovingIndex].corrSB.setValue(newCorr)
        # reset the moving state
        self._cpointMovingIndex = None
        taurusplot.canvas().setCursor(Qt.Qt.CrossCursor)

    def plot1MouseDoubleClickEvent(self, event):
        self.plotMouseDoubleClickEvent(event, self.plot1)

    def plot2MouseDoubleClickEvent(self, event):
        self.plotMouseDoubleClickEvent(event, self.plot2)

    def plotMouseDoubleClickEvent(self, event, taurusplot):
        picked, pickedCurveName, pickedIndex = taurusplot.pickDataPoint(
            event.pos(), scope=20, showMarker=False, targetCurveNames=['Control Points'])
        if picked is not None:
            return  # we dont want to create a control point too close of an existing one
        xp = taurusplot.invTransform(taurusplot.getCurve(
            'Control Points').xAxis(), event.x())
        if xp < self.xp[0] or xp > self.xp[-1]:
            return  # we dont want to create control points out of the curve range
        if Qt.QMessageBox.question(self, 'Create Control Point?', 'Insert a new control point at x=%g?' % xp, 'Yes', 'No') == 0:
            self.insertController(xp)
            self.changeCPointSelection(xp)
            # singleshot is used as a hack to get out of the eventhandler
            Qt.QTimer.singleShot(1, self.makeControllerVisible)

    def makeControllerVisible(self, ctrl=None):
        if ctrl is None:
            ctrl = self._selectedController
        self.scrollArea.ensureWidgetVisible(ctrl)

    def connectToController(self, ctrl):
        ctrl.selected.connect(self.changeCPointSelection)
        ctrl.corrSB.valueChanged.connect(self.onCorrSBChanged)
        ctrl.lCopyBT.clicked.connect(self.onLCopy)
        ctrl.rCopyBT.clicked.connect(self.onRCopy)
        ctrl.lScaleBT.clicked.connect(self.onLScale)
        ctrl.rScaleBT.clicked.connect(self.onRScale)

    def onAddSingleCPointBT(self):
        x = self._addCPointsDialog.singleCPointXSB.value()
        self.insertController(x)

    def onAddRegEspCPointsBT(self):
        cpoints = self._addCPointsDialog.regEspCPointsSB.value()
        positions = numpy.linspace(self.x[0], self.x[-1], cpoints + 2)[1:-1]
        for xp in positions:
            self.insertController(xp)

    def showEditCPointsDialog(self):
        dialog = EditCPointsDialog(self)
        table = dialog.tableTW
        table.setRowCount(self.xp.size)
        for i, (xp, corrp) in enumerate(zip(self.xp, self.corrp)):
            table.setItem(i, 0, Qt.QTableWidgetItem(str(xp)))
            table.setItem(i, 1, Qt.QTableWidgetItem(str(corrp)))

        # show dialog and update values if it is accepted
        if dialog.exec_():
            # delete previous controllers
            for c in self._controllers:
                c.setParent(None)
                c.deleteLater()
            self._controllers = []
            # and create them anew
            new_xp = numpy.zeros(table.rowCount())
            new_corrp = numpy.zeros(table.rowCount())
            try:
                for i in range(table.rowCount()):
                    new_xp[i] = float(table.item(i, 0).text())
                    new_corrp[i] = float(table.item(i, 1).text())
                self.setCorrection(new_xp, new_corrp)
            except:
                Qt.QMessageBox.warning(
                    self, 'Invalid data', 'Some values were not valid. Edition is ignored.')

    def _getInsertionIndex(self, xp):
        i = 0
        while (self.xp[i] < xp):
            i += 1
        return i

    def insertControllers(self, xplist):
        for xp in xplist:
            self.insertController(xp)

    def insertController(self, xp, index=None):
        if xp in self.xp:
            return None
        if index is None:
            index = self._getInsertionIndex(xp)
        # updating data (not in the most efficient way, but at least is clean)
        old_xp = self.xp
        self.xp = numpy.concatenate((self.xp[:index], (xp,), self.xp[index:]))
        self.yp = numpy.interp(self.xp, self.x, self.y)
        # the new correction is obtained by interpolation from the neighbouring
        # ones
        self.corrp = numpy.interp(self.xp, old_xp, self.corrp)
        # creating the controller
        ctrl = ControllerBox(parent=None, x=xp, y=self.yp[
                             index], corr=self.corrp[index])
        self.ctrlLayout.insertWidget(index, ctrl)
        self._controllers.insert(index, ctrl)
        self.connectToController(ctrl)
        self.updatePlots()
        return index

    def delController(self, index):
        c = self._controllers.pop(index)
        c.setParent(None)
        c.deleteLater()
        self.xp = numpy.concatenate((self.xp[:index], self.xp[index + 1:]))
        self.yp = numpy.interp(self.xp, self.x, self.y)
        self.corrp = numpy.concatenate(
            (self.corrp[:index], self.corrp[index + 1:]))

    def _populatePlots(self):
        # Curves appearance
        self._yAppearance = CurveAppearanceProperties(
            sStyle=Qwt5.QwtSymbol.NoSymbol,
            lStyle=Qt.Qt.SolidLine,
            lWidth=2,
            lColor='black',
            cStyle=Qwt5.QwtPlotCurve.Lines,
            yAxis=Qwt5.QwtPlot.yLeft)

        self._correctedAppearance = CurveAppearanceProperties(
            sStyle=Qwt5.QwtSymbol.NoSymbol,
            lStyle=Qt.Qt.DashLine,
            lWidth=1,
            lColor='blue',
            cStyle=Qwt5.QwtPlotCurve.Lines,
            yAxis=Qwt5.QwtPlot.yLeft)

        self._cpointsAppearance = CurveAppearanceProperties(
            sStyle=Qwt5.QwtSymbol.Rect,
            sSize=5,
            sColor='blue',
            sFill=True,
            lStyle=Qt.Qt.NoPen,
            yAxis=Qwt5.QwtPlot.yLeft)

        self._corrAppearance = CurveAppearanceProperties(
            sStyle=Qwt5.QwtSymbol.NoSymbol,
            lStyle=Qt.Qt.SolidLine,
            lWidth=1,
            lColor='blue',
            cStyle=Qwt5.QwtPlotCurve.Lines,
            yAxis=Qwt5.QwtPlot.yLeft)

        self.plot1.attachRawData({'x': self.x, 'y': self.y, 'title': 'Master'})
        self.plot1.setCurveAppearanceProperties({'Master': self._yAppearance})

        self.plot1.attachRawData(
            {'x': self.xp, 'y': self.yp + self.corrp, 'title': 'Control Points'})
        self.plot1.setCurveAppearanceProperties(
            {'Control Points': self._cpointsAppearance})

        self.plot1.attachRawData(
            {'x': self.x, 'y': self.y + self.corr, 'title': 'Corrected'})
        self.plot1.setCurveAppearanceProperties(
            {'Corrected': self._correctedAppearance})

        self.plot2.attachRawData(
            {'x': self.x, 'y': self.corr, 'title': 'Correction'})
        self.plot2.setCurveAppearanceProperties(
            {'Correction': self._corrAppearance})

        self.plot2.attachRawData(
            {'x': self.xp, 'y': self.corrp, 'title': 'Control Points'})
        self.plot2.setCurveAppearanceProperties(
            {'Control Points': self._cpointsAppearance})

    def updatePlots(self):
        self.plot1.getCurve('Control Points').setData(
            self.xp, self.yp + self.corrp)
        self.plot1.getCurve('Corrected').setData(self.x, self.y + self.corr)
        self.plot2.getCurve('Correction').setData(self.x, self.corr)
        self.plot2.getCurve('Control Points').setData(self.xp, self.corrp)
        index = self._getInsertionIndex(self.markerPos)
        self.marker1.setValue(self.xp[index], self.yp[
                              index] + self.corrp[index])
        self.marker2.setValue(self.xp[index], self.corrp[index])
        self.plot1.replot()
        self.plot2.replot()

    def onLCopy(self, checked):
        sender = self.sender().ctrlObj
        index = self._getInsertionIndex(sender._x)
        v = sender.corrSB.value()
        for ctrl in self._controllers[:index]:
            ctrl.corrSB.setValue(v)

    def onRCopy(self, checked):
        sender = self.sender().ctrlObj
        index = self._getInsertionIndex(sender._x)
        v = sender.corrSB.value()
        for ctrl in self._controllers[index + 1:]:
            ctrl.corrSB.setValue(v)

    def onLScale(self, checked):
        sender = self.sender().ctrlObj
        index = self._getInsertionIndex(sender._x)
        # y=numpy.interp(self.xp, self.x, self.y) #values of the master at the
        # control points
        if self.yp[index] == 0:
            Qt.QMessageBox.warning(
                self, 'Scaling Error', 'The master at this control point is zero-valued. This point cannot be used as reference for scaling')
            return
        v = sender.corrSB.value() / (self.yp[index])
        for i in range(0, index):
            self._controllers[i].corrSB.setValue(v * self.yp[i])

    def onRScale(self, checked):
        sender = self.sender().ctrlObj
        index = self._getInsertionIndex(sender._x)
        # y=numpy.interp(self.xp, self.x, self.y) #values of the master at the
        # control points
        if self.yp[index] == 0:
            Qt.QMessageBox.warning(
                self, 'Scaling Error', 'The master at this control point is zero-valued. This point cannot be used as reference for scaling')
            return
        v = sender.corrSB.value() / (self.yp[index])
        for i in range(index + 1, self.xp.size):
            self._controllers[i].corrSB.setValue(v * self.yp[i])

    def changeCPointSelection(self, newpos):
        index = self._getInsertionIndex(newpos)
        old_index = self._getInsertionIndex(self.markerPos)
        self.markerPos = self.xp[index]
        self.marker1.setValue(self.xp[index], self.yp[
                              index] + self.corrp[index])
        self.marker2.setValue(self.xp[index], self.corrp[index])
        self.plot1.replot()
        self.plot2.replot()
        self._controllers[old_index].corrSB.setStyleSheet('')
        self._controllers[index].corrSB.setStyleSheet('background:lightgreen')
        self._selectedController = self._controllers[index]

    def onCorrSBChanged(self, value=None):
        '''recalculates the position and value of the control points (self.xp and self.corrp)
        as well as the correction curve (self.corr)'''
        ctrl = self.sender().ctrlObj
        self.corrp[self._getInsertionIndex(ctrl._x)] = value
        # recalculate the correction curve
        self.corr = numpy.interp(self.x, self.xp, self.corrp)
        self.updatePlots()

    def setMaster(self, x, y, keepCP=False, keepCorr=False):
        # make sure that x,y are numpy arrays and that the values are sorted
        # for x
        x, y = numpy.array(x), numpy.array(y)
        if x.shape != y.shape or x.size == 0 or y.size == 0:
            raise ValueError('The master curve is not valid')
        sortedindexes = numpy.argsort(x)
        self.x, self.y = x[sortedindexes], y[sortedindexes]
        self.plot1.getCurve('Master').setData(self.x, self.y)
        xp = None
        corrp = None
        if self.x[0] == self.xp[0] and self.x[-1] == self.x[-1]:
            if keepCP:
                xp = self.xp
            if keepCorr:
                corrp = self.corrp
        self.setCorrection(xp, corrp)
        self._addCPointsDialog.singleCPointXSB.setRange(self.x[0], self.x[-1])

    def getMaster(self):
        '''returns x,m where x and m are numpy arrays representing the
        abcissas and ordinates for the master, respectively'''
        return self.x.copy(), self.y.copy()

    def resetMaster(self):
        x = numpy.arange(256, dtype='double')
        y = numpy.zeros(256, dtype='double')
        self.setMaster(x, y)

    def getCorrected(self):
        '''returns x,c where x and c are numpy arrays representing the
        abcissas and ordinates for the corrected curve, respectively'''
        return self.x.copy(), self.y + self.corr

    def getCorrection(self):
        '''returns xp,cp where xp and cp are numpy arrays representing the
        abcissas and ordinates for the correction points, respectively'''
        return self.xp.copy(), self.corrp.copy()

    def setCorrection(self, xp=None, corrp=None):
        '''sets control points at the points specified by xp and with the
        values specified by corrp. Example::

            setCorrection([1,2,8,9], [0,0,0,0])

        would set 4 control points with initial value 0 at x=1, 2, 8 and 9s
        '''
        for c in self._controllers:
            c.setParent(None)  # destroy previous controllers
            c.deleteLater()
        self._controllers = []
        if xp is None:
            xp = numpy.array((self.x[0], self.x[-1]))
            corrp = numpy.zeros(2)
        if corrp is None:
            corrp = numpy.zeros(xp.size)
        # make sure that the extremes are there
        if xp[0] > self.x[0]:
            xp = numpy.concatenate(((self.x[0],), xp))
            corrp = numpy.concatenate(((self.corrp[0],), corrp))
        if xp[-1] < self.x[-1]:
            xp = numpy.concatenate((xp, (self.x[-1],)))
            corrp = numpy.concatenate((corrp, (self.corrp[-1],)))

        # now create everything
        # make sure there are no repetitions and that it is sorted
        self.xp = numpy.unique(xp)
        # in case of repeated xp, take only one corrp
        self.corrp = numpy.interp(self.xp, xp, corrp)
        self.yp = numpy.interp(self.xp, self.x, self.y)
        for i, (x, c) in enumerate(zip(xp, corrp)):
            ctrl = ControllerBox(parent=None, x=xp[i], y=self.yp[i])
            self.ctrlLayout.insertWidget(i, ctrl)
            self._controllers.insert(i, ctrl)
            self.connectToController(ctrl)
        # recalculate the correction curve
        self.corr = numpy.interp(self.x, self.xp, self.corrp)
        self.updatePlots()
        self.changeCPointSelection(self.xp[0])

    def resetCorrection(self):
        self.setCorrection()
        #self.xp = numpy.linspace(self.x[0],self.x[-1], self.cpoints)
        #self.corrp = numpy.zeros(self.cpoints)
        # self._populateControllers()


if __name__ == "__main__":
    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)
    form = ArrayEditor()
    #x = numpy.arange(100)-20
    #y = -(x-50)**2+50**2
    x = numpy.linspace(0.1, 0.9, 100)
    y = (x) ** 2 - 5 * x
    form.setMaster(x, y)
    # form.setCorrection([1,30,70,90,100],[0,44,88,22,-100])
    form.show()
    # sys.exit(app.exec_())
    status = app.exec_()
    # x,y=form.getCorrected()
    # print "CORRECTED:",x,y
    sys.exit(status)



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import click
from .taurustrend import TaurusTrend
import taurus.cli.common


x_axis_mode = click.option(
    "-x", "--x-axis-mode", "x_axis_mode",
    type=click.Choice(['t', 'n']),
    default='n',
    show_default=True,
    help=('X axis mode. "t" implies using a Date axis'
          + '"n" uses the regular axis'),
)

def buffer(default):
    o = click.option(
        '-b', '--buffer', 'max_buffer_size',
        type=int,
        default=default,
        show_default=True,
        help=("Maximum number of values to be stacked "
              + "(when reached, the oldest values will be "
              + "discarded)"),
        )
    return o


@click.group('qwt5')
def qwt5():
    """[DEPRECATED] use "taurus plot" or "taurus trend" instead"""
    pass


@qwt5.command('plot')
@taurus.cli.common.models
@taurus.cli.common.config_file
@x_axis_mode
@taurus.cli.common.demo
@taurus.cli.common.window_name("TaurusPlot (qwt5)")
def plot_cmd(models, config_file, x_axis_mode, demo, window_name):
    """[DEPRECATED] use "taurus plot" instead"""
    from .taurusplot import plot_main
    return plot_main(models=models,
                     config_file=config_file,
                     x_axis_mode=x_axis_mode,
                     demo=demo,
                     window_name=window_name
                     )


@qwt5.command('trend')
@taurus.cli.common.models
@x_axis_mode
@taurus.cli.common.config_file
@taurus.cli.common.demo
@taurus.cli.common.window_name('TaurusTrend(qwt5)')
@buffer(TaurusTrend.DEFAULT_MAX_BUFFER_SIZE)
@click.option('-a', '--use-archiving', 'use_archiving',
              is_flag=True,
              default=False,
              help='enable automatic archiving queries')
@click.option('-r', '--forced-read', 'forced_read_period', type=int,
              default=-1,
              metavar="MILLISECONDS",
              help="force re-reading of the attributes every MILLISECONDS ms")
def trend_cmd(models, x_axis_mode, config_file, demo, window_name,
              max_buffer_size, use_archiving, forced_read_period):
    """[DEPRECATED] use "taurus trend" instead"""
    from .taurustrend import trend_main
    return trend_main(models=models,
                      config_file=config_file,
                      x_axis_mode=x_axis_mode,
                      use_archiving=use_archiving,
                      max_buffer_size=max_buffer_size,
                      forced_read_period=forced_read_period,
                      demo=demo,
                      window_name=window_name
                      )


if __name__ == '__main__':
    qwt5()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module provides widgets that display the database in a tree format"""

__all__ = [
    'getPixmap',
    'getIcon',
    'getThemePixmap',
    'getThemeIcon',
    'getThemeMembers',
    # the following are actually from taurus.qt.qtgui.icon
    'getStandardIcon',
    'getElementTypeToolTip',
    'getElementTypeSize',
    'getElementTypeIcon',
    'getElementTypeIconName',
    'getElementTypePixmap',
    'getDevStateToolTip',
    'getDevStateIcon',
    'getDevStatePixmap',]

__docformat__ = 'restructuredtext'

import os

from taurus.external.qt import Qt

from taurus.core.util.log import (deprecated, taurus4_deprecation,
                                  deprecation_decorator)

from taurus.qt.qtgui.icon import *


@deprecation_decorator(alt='QIcon.hasThemeIcon to test individual names',
                       rel='4.0')
def getThemeMembers():
    """Returns the current icon theme elements

    .. note:: Since its depredation, it returns an empty dict (there is no
    reasonable way of introspecting the list of available icon names).
    Alternatively Just test a given name using

    :return: the current icon theme elements in a dictionary where each key is
             a group name and the value is a sequence of theme icon name.
    :rtype: dict<str,seq<str>>"""
    return {}


def getPixmap(key, size=None):
    # handle resource syntax (deprecated)
    if key.startswith(':'):
        head, tail = os.path.split(key[1:])
        # logos used to be in the resource root. Now they are in 'logos'
        prefix = sanitizePrefix(head or 'logos')
        alt = 'getCachedPixmap("%s:%s [, size]")' % (prefix, tail)
        ret = getCachedPixmap('%s:%s' % (prefix, tail), size)
    deprecated(dep='getPixmap("%s" [, size])' % key, alt=alt, rel='4.0')
    return ret


def getIcon(key):
    """Returns a PyQt4.QtGui.QIcon object for the given key. It supports QDir's
    searchPath prefixes (see :meth:`QDir.setSearchPaths`).
    Note that taurus.qt.qtgui.resource already sets several search paths based
    on .path files

    :param key: (str) the pixmap file name. (optionally with a prefix)

    :return: (PyQt4.QtGui.QIcon) a PyQt4.QtGui.QIcon for the given key"""

    # handle resource syntax (deprecated)
    if key.startswith(':'):
        head, tail = os.path.split(key[1:])
        # logos used to be in the resource root. Now they are in 'logos'
        prefix = sanitizePrefix(head or 'logos')
        alt = 'Qt.QIcon("%s:%s")' % (prefix, tail)
        ret = Qt.QIcon('%s:%s' % (prefix, tail))
    elif not Qt.QFile.exists(key) and Qt.QIcon.hasThemeIcon(key):
        alt = 'QIcon.fromTheme("%s")' % key
        ret = Qt.QIcon.fromTheme(key)
    else:
        alt = 'QIcon("%s")' % key
        ret = Qt.QIcon(key)
    deprecated(dep='getIcon("%s")' % key, alt=alt, rel='4.0')
    return ret


@deprecation_decorator(alt='QIcon.fromTheme(key).pixmap(size, size)', rel='4.0')
def getThemePixmap(key, size=48):
    """Returns a PyQt4.QtGui.QPixmap object for the given key and size.
    Key should be a valid theme icon key. See :meth:`PyQt4.QIcon.fromTheme`.

    Note that if the OS does not define a theme, taurus.qt.qtgui.resource will
    use the bundled 'Tango' icons theme. See:
    `Tango Icon Library <http://tango.freedesktop.org/Tango_Icon_Library>`_.

    If the key cannot be found, it will return a null content Pixmap.

    :param key: (str) a string with the pixmap theme key (ex.: 'folder_open')
    :param size: (int) the pixmap size in pixels (will get a square pixmap).
                 Default size=48

    :return: (PyQt4.QtGui.QPixmap)
    """
    return Qt.QIcon.fromTheme(key).pixmap(size, size)


@deprecation_decorator(alt='QIcon.fromTheme', rel='4.0')
def getThemeIcon(key):
    """Returns the theme icon corresponding to the given key.
    Key should be a valid theme icon key. See :meth:`PyQt4.QIcon.fromTheme`.

    Note that if the OS does not define a theme, taurus.qt.qtgui.resource will
    use the bundled 'Tango' icons theme. See:
    `Tango Icon Library <http://tango.freedesktop.org/Tango_Icon_Library>`_.

    :param key: (str) a string with the icon theme key (e.g.: 'folder_open')

    :return: (PyQt4.QtGui.QIcon) a PyQt4.QtGui.QIcon for the given theme key
    """
    return Qt.QIcon.fromTheme(key)


@taurus4_deprecation(alt='getDevStateToolTip')
def getSWDevHealthToolTip(state):
    return getDevStateToolTip(state)


@taurus4_deprecation(alt='getDevStateIcon')
def getSWDevHealthIcon(state, fallback=None):
    return getDevStateIcon(state, fallback=fallback)


@taurus4_deprecation(alt='getDevStatePixmap')
def getSWDevHealthPixmap(state, fallback=None):
    return getDevStatePixmap(state, fallback=fallback)




if __name__ == '__main__':

    import sys
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(cmd_line_parser=None)



    themekey = 'computer'
    b = Qt.QIcon.fromTheme(themekey)

    icons = [

        getIcon('actions:edit-cut.svg'),
        getIcon(':/actions/edit-cut.svg'),

        getIcon(":/apps/preferences-system-session.svg"),
        getIcon(":/designer/devs_tree.png"),

        getIcon(":/actions/process-stop.svg"), # from tango-icons/actions
        getIcon(":/actions/add.svg"), # from rrze-icons/actions
        getIcon(":/actions/stop.svg"), # from extra-icons/actions

        getIcon(":taurus.svg"),
        getIcon("computer"), # theme Icon via getIcon
        getThemeIcon("computer"), # theme Icon via getThemeIcon
        ]

    w = Qt.QWidget()
    l = Qt.QVBoxLayout()
    w.setLayout(l)
    for icon in icons:
        button = Qt.QPushButton(icon, 'kk')
        l.addWidget(button)

    w.show()

    sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains color codes for state and quality"""
from __future__ import print_function
from builtins import object

__all__ = ["DEVICE_STATE_DATA", "ATTRIBUTE_QUALITY_DATA", "ColorPalette",
           "DEVICE_STATE_PALETTE", "ATTRIBUTE_QUALITY_PALETTE"]


DEVICE_STATE_DATA = {
    # map for TaurusDevState states (used for agnostic TaurusDevice.state)
    "TaurusDevState.Ready": ("Green", 0, 255, 0, 0),
    "TaurusDevState.NotReady": ("Red", 255, 0, 0, 0),
    "TaurusDevState.Undefined": ("Gray", 128, 128, 128, 0),

    # The following is a bck-compat map for taurus.core.tango.DevState
    # (i.e. PyTango.DevState). Used for "state" TangoAttribute rvalues
    "ON": ("Dead Frog Green", 0, 255, 0, 0),
    "OFF": ('White', 255, 255, 255, 0),
    "CLOSE": ("White", 255, 255, 255, 3),
    "OPEN": ("Green", 0, 255, 0, 0),
    "INSERT": ("White", 255, 255, 255, 0),
    "EXTRACT": ("Green", 0, 255, 0, 0),
    "MOVING": ("Light Blue", 128, 160, 255, 0),
    "STANDBY": ("Yellow", 255, 255, 0, 0),
    "FAULT": ("Red", 255, 0, 0, 0),
    "INIT": ("Grenoble", 204, 204, 122, 0),
    "RUNNING": ("Light Blue", 128, 160, 255, 0),
    "ALARM": ("Tangorange", 255, 140,   0, 1),
    "DISABLE": ("Magenta", 255,   0, 255, 0),
    "UNKNOWN": ("Gray", 128, 128, 128, 0),
    str(None): ("Gray", 128, 128, 128, 0),
    # Support also explicit keys from str representations of
    # taurus.core.tango.util.enums.DevState
    "DevState.ON": ("Dead Frog Green", 0, 255, 0, 0),
    "DevState.OFF": ('White', 255, 255, 255, 0),
    "DevState.CLOSE": ("White", 255, 255, 255, 3),
    "DevState.OPEN": ("Green", 0, 255, 0, 0),
    "DevState.INSERT": ("White", 255, 255, 255, 0),
    "DevState.EXTRACT": ("Green", 0, 255, 0, 0),
    "DevState.MOVING": ("Light Blue", 128, 160, 255, 0),
    "DevState.STANDBY": ("Yellow", 255, 255, 0, 0),
    "DevState.FAULT": ("Red", 255, 0, 0, 0),
    "DevState.INIT": ("Grenoble", 204, 204, 122, 0),
    "DevState.RUNNING": ("Light Blue", 128, 160, 255, 0),
    "DevState.ALARM": ("Tangorange", 255, 140,   0, 1),
    "DevState.DISABLE": ("Magenta", 255,   0, 255, 0),
    "DevState.UNKNOWN": ("Gray", 128, 128, 128, 0),
}

ATTRIBUTE_QUALITY_DATA = {
    "ATTR_INVALID": ("Gray", 128, 128, 128, 1),
    "ATTR_VALID": ("Frog Green", 0, 255, 0, 0),
    "ATTR_ALARM": ("Orange", 255, 140, 0, 1),
    "ATTR_WARNING": ("Orange", 255, 140, 0, 1),
    "ATTR_CHANGING": ("Lightblue", 128, 160, 255, 0),
    "UNKNOWN": ("Gray", 128, 128, 128, 0),
    str(None): ("Gray", 128, 128, 128, 0),
}

_BW_RGB = [(0, 0, 0, "Black"), (255, 255, 255, "White"),
           (255, 255, 0, "Yellow"), (0, 128, 0, "Green")]


class ColorPalette(object):
    """Provides the list of taurus colors equivalent to Tango colors."""

    def __init__(self, dat, int_decoder_dict=None):

        self._rgb_data = dat
        self._int_decoder_dict = int_decoder_dict

    def _decoder(self, elem):
        if type(elem) == int:
            elem = self._int_decoder_dict.get(elem)
        return str(elem)

    def rgb(self, stoq, fg=False):
        """Returns a triplet of rgb colors in the range of 0 .. 255."""
        name = self._decoder(stoq)
        dat = self._rgb_data[name]
        if fg:
            return _BW_RGB[dat[4]][0:3]
        else:
            return dat[1:4]

    def rgb_pair(self, stoq):
        """Returns pair of foreground and background colors."""
        return (self.rgb(stoq), self.rgb(stoq, fg=True))

    def hex(self, stoq, fmt="%06x", fg=False):
        """Returns the rgb colors as string of lowercase hexadecimal characters"""
        return fmt % self.number(stoq, fg)

    def number(self, stoq, fg=False):
        """Returns the colors as a number,
        suitable for conversion to hexadecimal as argument to QtGui.QColor."""
        r = self.rgb(stoq, fg)
        return r[0] * 256 * 256 + r[1] * 256 + r[2]

    def __iter__(self):
        return list(self._rgb_data.keys()).__iter__()

    def name(self, stoq, fg=False):
        """Returns the name of the color."""
        name = self._decoder(stoq)
        if fg:
            return _BW_RGB[self._rgb_data[name][4]][3]
        else:
            return self._rgb_data[name][0]

    def has(self, name):
        return name in self._rgb_data

    def size(self):
        return len(self._rgb_data)

    def format_SimStates(self, var="T1"):
        count = DEVICE_STATE_PALETTE.size() - 1
        txt = ""
        for state in DEVICE_STATE_PALETTE:
            txt += "%s=Attr('%s')>%d\n" % (state, var, count)
            count = count - 1
        return txt

    def htmlStyle(self, htmlTag, stoq):
        name = self._decoder(stoq)
        bg = str(self.rgb(stoq))
        fg = str(self.rgb(stoq, fg=True))
        txt = """<style type='text/css'>
%s.%s { background-color : rgb%s;  color : rgb%s; }
</style>""" % (htmlTag, name, bg, fg)
        return txt

    def qtStyleSheet(self, stoq):
        name = self._decoder(stoq)
        bg = str(self.rgb(stoq))
        fg = str(self.rgb(stoq, fg=True))
        txt = "background-color : rgb%s;  color : rgb%s; " % (bg, fg)
        return txt


class _DeprecationDecoder(list):
    def __init__(self, palette, elements):
        self.palette = palette
        list.__init__(self, elements)

    def get(self, i):
        from taurus.core.util.log import deprecated
        deprecated(dep='Using ints for accessing elements of %s' % self.palette,
                   alt='"%s"' % self[i], rel='4.0')
        return self[i]


_PYTANGO_DEVSTATE_INT_DECODER = _DeprecationDecoder(
        'DEVICE_STATE_PALETTE',
        ["ON",
         "OFF",
         "CLOSE",
         "OPEN",
         "INSERT",
         "EXTRACT",
         "MOVING",
         "STANDBY",
         "FAULT",
         "INIT",
         "RUNNING",
         "ALARM",
         "DISABLE",
         "UNKNOWN",])

_PYTANGO_ATTRQUALITY_INT_DECODER = _DeprecationDecoder(
        'ATTRIBUTE_QUALITY_PALETTE',
        ["ATTR_VALID",
         "ATTR_INVALID",
         "ATTR_ALARM",
         "ATTR_CHANGING",
         "ATTR_WARNING",])


DEVICE_STATE_PALETTE = ColorPalette(DEVICE_STATE_DATA,
                                    _PYTANGO_DEVSTATE_INT_DECODER)

ATTRIBUTE_QUALITY_PALETTE = ColorPalette(ATTRIBUTE_QUALITY_DATA,
                                         _PYTANGO_ATTRQUALITY_INT_DECODER)


def print_color_palette(pal):
    """Prints a list of colors to stdout."""
    for stoq in pal:
        fg_color = pal.name(stoq, fg=True)
        bg_color = pal.name(stoq)
        rgb = "(%3.3d, %3.3d, %3.3d)" % pal.rgb(stoq)
        hx = pal.hex(stoq)
        print("%7s %5s on %13s %15s #%s" % (stoq, fg_color, bg_color, rgb, hx))


if __name__ == "__main__":
    print_color_palette(DEVICE_STATE_PALETTE)
    print_color_palette(ATTRIBUTE_QUALITY_PALETTE)
    from taurus.core import TaurusDevState
    import PyTango
    print()
    print(DEVICE_STATE_PALETTE.rgb(TaurusDevState.Ready))
    print(DEVICE_STATE_PALETTE.rgb('TaurusDevState.Ready'))
    print(DEVICE_STATE_PALETTE.rgb(PyTango.DevState.ON))
    print(DEVICE_STATE_PALETTE.rgb(0))



import os
import sys
import importlib
import pytest
from taurus import tauruscustomsettings as ts


@pytest.mark.parametrize(
    "qt_api, default_qt_api, installed, imported, expected",
    [
        # no explicit selection, all bindings installed
        ("", "", "all", None, "pyqt5"),
        (None, "", "all", None, "pyqt5"),
        # no explicit selection, only one binding installed
        (None, "", "PyQt5", None, "pyqt5"),
        (None, "", "PyQt4", None, "pyqt"),
        (None, "", "PySide2", None, "pyside2"),
        (None, "", "PySide", None, "pyside"),
        # no explicit selection, with default selection, all installed
        (None, "pyqt5", "all", None, "pyqt5"),
        (None, "pyqt", "all", None, "pyqt"),
        (None, "pyqt4", "all", None, "pyqt"),
        (None, "pyside2", "all", None, "pyside2"),
        (None, "pyside", "all", None, "pyside"),
        # explicit selection, all bindings installed
        ("pyqt5", "", "all", None, "pyqt5"),
        ("pyqt", "", "all", None, "pyqt"),
        ("pyqt4", "", "all", None, "pyqt"),
        ("pyside2", "", "all", None, "pyside2"),
        ("pyside", "", "all", None, "pyside"),
        ("pyqt5", "pyqt4", "all", None, "pyqt5"),
        ("pyqt", "pyqt5", "all", None, "pyqt"),
        ("pyqt4", "pyqt5", "all", None, "pyqt"),
        # unsupported selection
        ("unsupported_binding", "", "all", None, ImportError),
        (None, "unsupported_binding", "all", None, ImportError),
        ("pyqt5", "", "none", None, ImportError),
        ("pyqt5", "", "PyQt4", None, ImportError),
        ("pyqt", "", "PyQt5", None, ImportError),
        ("pyside2", "", "PySide", None, ImportError),
        ("pyside", "", "PySide2", None, ImportError),
        # previously imported binding
        (None, "", "all", ("PyQt5",), "pyqt5"),
        (None, "", "all", ("PyQt4",), "pyqt"),
        (None, "", "all", ("PySide2",), "pyside2"),
        (None, "", "all", ("PySide",), "pyside"),
        (None, "", "all", ("PySide.QtCore",), "pyside"),
        (None, "", "all", (".QtCore", "PyQt4"), "pyqt"),
        ("pyqt5", "pyqt4", "all", ("PySide2",), "pyside2"),
    ],
)
@pytest.mark.forked  # run in separate process to avoid side-effects
def test_qt_select(
    monkeypatch, qt_api, default_qt_api, installed, imported, expected
):
    """Check that the selection of Qt binding by taurus.external.qt works"""
    # temporarily remove qt bindings from sys.modules
    monkeypatch.delitem(sys.modules, "taurus.external.qt", raising=False)
    for binding in "PyQt5", "PyQt4", "PySide2", "PySide":
        monkeypatch.delitem(sys.modules, binding, raising=False)
        monkeypatch.delitem(sys.modules, binding + ".QtCore", raising=False)
    # monkeypatch QT_API and DEFAULT_QT_API with the values from arguments
    if qt_api is None:
        monkeypatch.delenv("QT_API", raising=False)
    else:
        monkeypatch.setenv("QT_API", qt_api)
    monkeypatch.setattr(ts, "DEFAULT_QT_API", default_qt_api)
    # avoid initializations
    monkeypatch.setattr(ts, "QT_AUTO_INIT_LOG", False)
    monkeypatch.setattr(ts, "QT_AUTO_REMOVE_INPUTHOOK", False)
    monkeypatch.setattr(ts, "QT_AVOID_ABORT_ON_EXCEPTION", False)
    # provide importable mocks for all supported bindings
    monkeypatch.syspath_prepend(os.path.join(os.path.dirname(__file__), "res"))
    monkeypatch.setenv("AVAILABLE_QT_MOCKS", installed)
    # emulate an already-imported binding
    if imported is not None:
        importlib.import_module(*imported)
    # Now that the environment is clean and ready, test the shim
    if not isinstance(expected, str):
        with pytest.raises(expected):
            from taurus.external.qt import API
    else:
        from taurus.external.qt import API

        assert API == expected



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import click
import sys

@click.group('guiqwt')
def guiqwt():
    """[DEPRECATED] use "taurus image" or "taurus trend2d" instead"""
    print('"taurus guiqwt" subcommand is deprecated. '
          + 'Use "taurus image" or "taurus trend2d" instead\n')
    sys.exit(1)


@guiqwt.command()
def image():
    """[DEPRECATED] use "taurus image" instead"""
    sys.exit(1)


@guiqwt.command()
def trend2d():
    """[DEPRECATED] use "taurus trend2d" instead"""
    sys.exit(1)


if __name__ == '__main__':
    guiqwt()




MyAttr = 'eval:1'
My_Attr = 'eval:foo=1;bar=2;foo+bar'
attr_1 = 'tango:a/b/c/d'
attr_2 = 'a/b/c/d'
attr1 = 'eval:"Hello_World!!"'
foo = 'eval:/@Foo/True'
# 1foo = 'eval:2'
Foo = 'eval:False'
res_attr = 'res:attr1'
dev1 = 'tango:a/b/c'  # invalid for attribute
dev2 = 'eval:@foo'   # invalid for attribute



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
properties.py: Methods for adding QProperties to QObjects

A call like
        set_property_methods(self,'Filters','QString',default='',
            set_callback=lambda s=self:s.loadTree(s.getFilters(),clear=True),
            reset_callback=lambda s=self:s.loadTree('',clear=True)
            )

Would replace all these lines:

    def setFilters(self,filters):
        self._filters = filters
        self.loadTree(self._filters,clear=True)

    def getFilters(self):
        return self._filters

    def resetFilters(self):
        self._filters=""
        self.loadTree(self._filters)

    filters = QtCore.pyqtProperty("QString", getFilters, setFilters, resetFilters)

Not tested yet with the classical declaration:

    #model = QtCore.pyqtProperty("QString", TaurusBaseWidget.getModel,
                            #TaurusBaseWidget.setModel,
                            #TaurusBaseWidget.resetModel)

"""

from builtins import str
from builtins import map
from taurus.external.qt import Qt
from taurus.core.util.fandango_search import isSequence, isDictionary


def join(*seqs):
    """ It returns a list containing the objects of all given sequences. """
    if len(seqs) == 1 and isSequence(seqs[0]):
        seqs = seqs[0]
    result = []
    for seq in seqs:
        if isSequence(seq):
            result.extend(seq)
        else:
            result.append(seq)
    #    result += list(seq)
    return result


def djoin(a, b):
    """ This method merges dictionaries and/or lists """
    if not any(map(isDictionary, (a, b))):
        return join(a, b)
    other, dct = sorted((a, b), key=isDictionary)
    if not isDictionary(other):
        other = dict.fromkeys(other if isSequence(other) else [other, ])
    for k, v in other.items():
        dct[k] = v if not k in dct else djoin(dct[k], v)
    return dct


def get_property_attribute(name):
    return '_' + str(name).lower()


def get_property(obj, name, callback=None):
    return (callback and callback()) or getattr(obj, get_property_attribute(name))


def set_property(obj, name, value, callback=None):
    # print 'set_property(%s,%s,%s,%s)'%(obj,name,value,callback)
    setattr(obj, get_property_attribute(name), value)
    try:
        callback and callback(value)
    except:
        callback()


def reset_property(obj, name, default=None, callback=None):
    setattr(obj, get_property_attribute(name), default)
    if callback:
        callback()

COMMON_PROPERTIES = (
    'ModelInConfig',  # If True, it will automatically register Model
    'modifiableByUser',
    #'useParentModel', #Only for widgets, not for components
    #'Model', #Controlled by ModelInConfig
)


def set_property_methods(obj, name, type_="QString", default=None, getter=None,
                         setter=None, reset=None, get_callback=None,
                         set_callback=None, reset_callback=None, qt=False,
                         config=False):
    """
    This method allows to add QProperties dynamically with calls like::

        set_property_methods(self,'Filters','QString',default='',
            set_callback=lambda s=self:s.loadTree(s.getFilters(),clear=True),
            reset_callback=lambda s=self:s.loadTree('',clear=True)
            )


    .. todo: This method should be refactored using python
             descriptors/properties and types.MethodType
    """
    klass = obj.__class__
    mname = '%s%s' % (name[0].upper(), name[1:])
    lname = '%s%s' % (name[0].lower(), name[1:])
    getter = getter or (lambda o=obj, n=name, c=get_callback: get_property(
        o, n, c))  # partial(get_property,name=name,callback=get_callback)
    setter = setter or (lambda x, y=None, o=obj, d=default, n=name, c=set_callback: set_property(
        o, n, x if x is not obj else y, c))  # partial(set_property),name=name,callback=set_callback)
    reset = reset or (lambda o=obj, n=name, d=default, c=reset_callback: reset_property(
        o, n, d, c))  # partial(reset_property,name=name,default=default,callback=reset_callback)
    setattr(obj, 'set%s' % mname, setter)
    setattr(obj, 'get%s' % mname, getter)
    setattr(obj, 'reset%s' % mname, reset)
    if qt:
        setattr(klass, lname,
                Qt.pyqtProperty("QString", getter, setter, reset)
                )
    if config:
        obj.registerConfigProperty(getter, setter, name)
    reset()



# -*- coding: utf-8 -*-
#
# Copyright © 2018- CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright © 2014-2015 Colin Duquesnoy
# Copyright © 2009-2018 The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides QtCore classes and functions.
"""
from builtins import str as __str
from taurus.core.util.log import deprecation_decorator as __deprecation

from . import PYQT5, PYSIDE2, PYQT4, PYSIDE, PythonQtError


# --------------------------------------------------------------------------
# QString, from_qvariant and to_qvariant are kept for now to
# facilitate transition of existing code but using them
# should be avoided (they only make sense with API 1, which is not supported)
@__deprecation(rel='4.0.1', alt='str')
class QString(__str):
    pass


@__deprecation(rel='4.0.1', alt='python objects directly')
def from_qvariant(qobj=None, convfunc=None):
    return qobj


@__deprecation(rel='4.0.1', alt='python objects directly')
def to_qvariant(pyobj=None):
    return pyobj
# --------------------------------------------------------------------------

if PYQT5:
    from PyQt5.QtCore import *
    from PyQt5.QtCore import pyqtSignal as Signal
    from PyQt5.QtCore import pyqtSlot as Slot
    from PyQt5.QtCore import pyqtProperty as Property
    from PyQt5.QtCore import QT_VERSION_STR as __version__

    # For issue #153 of qtpy
    from PyQt5.QtCore import QDateTime
    QDateTime.toPython = QDateTime.toPyDateTime

elif PYSIDE2:
    from PySide2.QtCore import *
    from PySide2.QtCore import Signal as pyqtSignal
    from PySide2.QtCore import Slot as pyqtSlot

    # ------------------------------------------------
    # Calling Property with doc="" produces segfaults in the tests.
    # As a workaround, just remove the doc kwarg. Note this is an ad-hoc
    # workaround: I could not find the API definition for
    # PySide.QtCore.Property in order to do a more complete mock
    def pyqtProperty(*args, **kwargs):
        kwargs.pop('doc', None)
        return Property(*args, **kwargs)
    # -------------------------------------------------

    try:  # may be limited to PySide-5.11a1 only
        from PySide2.QtGui import QStringListModel
    except:
        pass

elif PYQT4:
    from PyQt4.QtCore import *
    from PyQt4.QtCore import QCoreApplication
    from PyQt4.QtCore import Qt
    from PyQt4.QtCore import pyqtSignal as Signal
    from PyQt4.QtCore import pyqtSlot as Slot
    from PyQt4.QtCore import pyqtProperty as Property
    from PyQt4.QtGui import (QItemSelection, QItemSelectionModel,
                             QItemSelectionRange, QSortFilterProxyModel,
                             QStringListModel)
    from PyQt4.QtCore import QT_VERSION_STR as __version__

    # QDesktopServices has has been split into (QDesktopServices and
    # QStandardPaths) in Qt5
    # This creates a dummy class that emulates QStandardPaths
    from PyQt4.QtGui import QDesktopServices as _QDesktopServices

    class QStandardPaths():
        StandardLocation = _QDesktopServices.StandardLocation
        displayName = _QDesktopServices.displayName
        DesktopLocation = _QDesktopServices.DesktopLocation
        DocumentsLocation = _QDesktopServices.DocumentsLocation
        FontsLocation = _QDesktopServices.FontsLocation
        ApplicationsLocation = _QDesktopServices.ApplicationsLocation
        MusicLocation = _QDesktopServices.MusicLocation
        MoviesLocation = _QDesktopServices.MoviesLocation
        PicturesLocation = _QDesktopServices.PicturesLocation
        TempLocation = _QDesktopServices.TempLocation
        HomeLocation = _QDesktopServices.HomeLocation
        DataLocation = _QDesktopServices.DataLocation
        CacheLocation = _QDesktopServices.CacheLocation
        writableLocation = _QDesktopServices.storageLocation


    # Deprecated. QVariant is kept for now to facilitate transition of existing
    # code but using QVariant should be avoided (with API 2 it is superfluous)
    @__deprecation(rel='4.0.1', alt='python objects directly')
    def QVariant(pyobj=None):
        return pyobj


elif PYSIDE:
    from PySide.QtCore import *
    from PySide.QtCore import Signal as pyqtSignal
    from PySide.QtCore import Slot as pyqtSlot
    from PySide.QtCore import Property as pyqtProperty
    from PySide.QtGui import (QItemSelection, QItemSelectionModel,
                              QItemSelectionRange, QSortFilterProxyModel,
                              QStringListModel)

    # QDesktopServices has has been split into (QDesktopServices and
    # QStandardPaths) in Qt5
    # This creates a dummy class that emulates QStandardPaths
    from PySide.QtGui import QDesktopServices as _QDesktopServices

    class QStandardPaths():
        StandardLocation = _QDesktopServices.StandardLocation
        displayName = _QDesktopServices.displayName
        DesktopLocation = _QDesktopServices.DesktopLocation
        DocumentsLocation = _QDesktopServices.DocumentsLocation
        FontsLocation = _QDesktopServices.FontsLocation
        ApplicationsLocation = _QDesktopServices.ApplicationsLocation
        MusicLocation = _QDesktopServices.MusicLocation
        MoviesLocation = _QDesktopServices.MoviesLocation
        PicturesLocation = _QDesktopServices.PicturesLocation
        TempLocation = _QDesktopServices.TempLocation
        HomeLocation = _QDesktopServices.HomeLocation
        DataLocation = _QDesktopServices.DataLocation
        CacheLocation = _QDesktopServices.CacheLocation
        writableLocation = _QDesktopServices.storageLocation

    import PySide.QtCore
    __version__ = PySide.QtCore.__version__
else:
    raise PythonQtError('No Qt bindings could be found')


#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Taurus Plot module (old)
========================

This module is now deprecated.

If available, it exposes the plot API now implemented in
`taurus.qt.qtgui.qwt5`. Otherwise it tries to provide a minimal API
(essentially the TaurusPlot and TaurusTrend classes from taurus_pyqtgraph)
to help with the transition to another plotting module such as
taurus.qt.qtgui.tpg.

Note that if you really want to continue using the old Qwt5-based widgets and
avoid deprecation warnings, you can import the old API from
taurus.qt.qtgui.qwt5

Note that `PyQwt5 module <http://pyqwt.sourceforge.net/>`_
only works with Python2 and PyQt4 and is no longer supported,
so taurus is moving to other modules for plotting (pyqtgraph, silx, ...)

"""

from __future__ import absolute_import
from taurus.core.util import log as __log

__log.deprecated(dep='taurus.qt.qtgui.plot', rel='4.5',
                 alt='taurus.qt.qtgui.tpg or taurus.qt.qtgui.qwt5')

try:
    # Import all from qwt5
    from taurus.qt.qtgui.qwt5 import *
    # ...and patch sys.modules to expose all qwt5 submodules here
    # (even those not in the public API). This fixes
    # https://github.com/taurus-org/taurus/issues/909)
    import sys
    d = {}
    for k, v in sys.modules.items():
        if 'taurus.qt.qtgui.qwt5.' in k:
            d[k.replace('taurus.qt.qtgui.qwt5.', 'taurus.qt.qtgui.plot.')] = v
    sys.modules.update(d)
except:
    try:
        from taurus.qt.qtgui.tpg import TaurusPlot, TaurusTrend
        __log.info('plot: Using taurus.qt.qtgui.tpg to provide a minimal API '
                   + 'to facilitate the transition')
    except:
        __log.info('plot: Cannot import taurus.qt.qtgui.tpg to provide a '
                   + 'minimal API for transition')






#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
tauruswidgetfactory.py:
"""


__all__ = ["TaurusWidgetFactory", "getWidgetsOfType"]

__docformat__ = 'restructuredtext'

import imp
import os.path

from taurus.external.qt import Qt

from taurus.core.util.log import Logger
from taurus.core.util.singleton import Singleton

import taurus.qt.qtgui.base


def _getWidgetsOfType(widget, widgets, class_or_type_or_tuple):

    if isinstance(widget, class_or_type_or_tuple):
        widgets.append(widget)
    for w in widget.children():
        if isinstance(w, Qt.QWidget):
            _getWidgetsOfType(w, widgets, class_or_type_or_tuple)


def getWidgetsOfType(widget, class_or_type_or_tuple):
    """Returns all widgets in a hierarchy of a certain type

    :param widget: the widget to be inspected
    :type widget: Qt.QWidget
    :param class-or-type-or-tuple: type to be checked
    :type class-or-type-or-tuple: type class or a tuple of type classes

    :return: a sequence containning all widgets in the hierarchy that match the
             given type
    :rtype: seq<Qt.QWidget>"""
    widgets = []
    _getWidgetsOfType(widget, widgets, class_or_type_or_tuple)
    return widgets


# TODO: TaurusWidgetFactory should be deprecated in favor of using importlib

class TaurusWidgetFactory(Singleton, Logger):
    """The TaurusWidgetFactory is a utility class that provides information
    about all Qt widgets (Taurus and non Taurus) that are found in the
    current taurus distribution.
    TaurusWidgetFactory is a singleton class so you can freely create new
    instances since they will all reference the same object.
    Usage::

        from taurus.qt.qtgui.util import TaurusWidgetFactory

        wf = TaurusWidgetFactory()
        print wf.getTaurusWidgetClassNames()"""

    skip_modules = ('widget', 'util', 'qtdesigner', 'uic', 'resource',
                    'console')

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""

    def init(self, *args):
        """Singleton instance initialization."""
        name = self.__class__.__name__
        self.call__init__(Logger, name)

        path = os.path.dirname(os.path.abspath(__file__))
        path, tail = os.path.split(path)
        self._taurus_widgets, self._qt_widgets = self._buildWidgets(
            'taurus.qt.qtgui', path)
        self._addExtraTaurusWidgets(self._taurus_widgets, self._qt_widgets)

    def _buildWidgets(self, module_name, path, recursive=True):
        import taurus.qt.qtgui.base

        if Qt.QApplication.instance() is None:
            app = Qt.QApplication([])

        elems = os.listdir(path)
        taurus_ret, qt_ret = {}, {}
        if not '__init__.py' in elems:
            return taurus_ret, qt_ret

        try:
            m = __import__(module_name, fromlist=['*'], level=0)
            dir_names = dir(m)
            for dir_name in dir_names:
                if dir_name.startswith("_"):
                    continue
                try:
                    attr = getattr(m, dir_name)
                    if issubclass(attr, Qt.QWidget):
                        package = m.__package__
                        qt_ret[dir_name] = package, attr
                        if issubclass(attr, taurus.qt.qtgui.base.TaurusBaseWidget):
                            taurus_ret[dir_name] = package, attr
                except Exception as e:
                    pass
        except Exception as e:
            return taurus_ret, qt_ret

        if not recursive:
            return taurus_ret, qt_ret

        for elem in elems:
            abs_elem = os.path.join(path, elem)
            if (not elem.startswith('.')) and os.path.isdir(abs_elem) and \
               (not elem in self.skip_modules):
                m_name = os.path.splitext(elem)[0]
                new_module_name = '%s.%s' % (module_name, m_name)
                new_taurus_ret, new_qt_ret = self._buildWidgets(
                    new_module_name, abs_elem, True)
                taurus_ret.update(new_taurus_ret)
                qt_ret.update(new_qt_ret)
        return taurus_ret, qt_ret

    def _addExtraTaurusWidgets(self, taurus_ret, qt_widgets):
        designer_path = os.environ.get('TAURUSQTDESIGNERPATH')
        if designer_path is None:
            return taurus_ret
        designer_path = designer_path.split(os.path.pathsep)
        for path in designer_path:
            self._addExtraTaurusWidgetsPath(taurus_ret, qt_widgets, path)

    def _addExtraTaurusWidgetsPath(self, taurus_ret, qt_widgets, path):
        self.debug("Trying extra taurus widgets in %s", path)
        path = os.path.abspath(path)
        if not os.path.isdir(path):
            return
        elems = os.listdir(path)
        for elem in elems:
            m_name, ext = os.path.splitext(elem)
            if ext != '.py':
                continue
            try:
                self.debug("Trying to find extra module %s", m_name)
                f, fname, data = imp.find_module(m_name, [path])
            except ImportError as ie:
                self.debug("Could not find extra module %s:%s", m_name, ie)
                continue
            try:
                self.debug("Trying to load extra module %s", m_name)
                mod = imp.load_module(m_name, f, fname, data)
            except ImportError as ie:
                self.debug("Could not load extra module %s:%s", m_name, ie)
                continue
            dir_names = dir(mod)
            for dir_name in dir_names:
                if dir_name.startswith("_"):
                    continue
                if dir_name in taurus_ret:
                    continue
                try:
                    attr = getattr(mod, dir_name)
                    if issubclass(attr, Qt.QWidget):
                        if issubclass(attr, taurus.qt.qtgui.base.TaurusBaseWidget):
                            qt_info = attr.getQtDesignerPluginInfo()
                            taurus_ret[dir_name] = qt_info['module'], attr
                            qt_widgets[dir_name] = qt_info['module'], attr
                            self.debug("registered taurus widget %s", dir_name)
                except Exception as e:
                    pass

    def getWidgets(self):
        return self._qt_widgets

    def getTaurusWidgets(self):
        return self._taurus_widgets

    def getWidgetClassNames(self):
        return list(self._qt_widgets.keys())

    def getWidgetClasses(self):
        return [klass for mod_name, klass in self._qt_widgets.values()]

    def getWidgetClass(self, name):
        return self._qt_widgets[name][1]

    def getTaurusWidgetClassNames(self):
        return list(self._taurus_widgets.keys())

    def getTaurusWidgetClasses(self):
        return [klass for mod_name, klass in self._taurus_widgets.values()]

    def getTaurusWidgetClass(self, name):
        return self._taurus_widgets.get(name)[1]



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module is designed to provide a factory class for taurus Qt actions """

from __future__ import absolute_import

from future.utils import string_types

from taurus.core.util.log import Logger
from taurus.core.util.singleton import Singleton
from taurus.external.qt import Qt

from . import taurusaction


__all__ = ["ActionFactory"]

__docformat__ = 'restructuredtext'


class ActionFactory(Singleton, Logger):
    """A Singleton class designed to provide Action related objects."""

    def __init__(self):
        """ Initialization. Nothing to be done here for now."""
        pass

    def init(self, *args):
        """Singleton instance initialization."""
        self.call__init__(Logger, 'ActionFactory')
        self.actions = self.__getActions()
        self.menus = self.__getMenus()

    def __getClasses(self, super_class):
        ret = {}
        klass_type = type(super_class)
        for name in dir(taurusaction):
            klass = getattr(taurusaction, name)
            if klass == super_class:
                continue
            if (type(klass) == klass_type and issubclass(klass, super_class)):
                ret[klass.menuID] = klass
        return ret

    def __getActions(self):
        """Calculates the map of existing action classes"""
        return self.__getClasses(taurusaction.TaurusAction)

    def __getMenus(self):
        """Calculates the map of existing menu classes"""
        return self.__getClasses(taurusaction.TaurusMenu)

    def getActions(self):
        return self.actions

    def getMenus(self):
        return self.menus

    def getNewAction(self, widget, id):
        klass = self.actions.get(id)
        if klass is None:
            return None
        return klass(widget)

    def getNewMenu(self, widget, data):
        import xml.dom.minidom
        doc = xml.dom.minidom.parseString(data)
        m_node = doc.childNodes[0]
        return self.buildMenu(widget, m_node)

    def buildAction(self, widget, a_node):
        if not a_node.hasAttribute('class'):
            return None

        id = a_node.getAttribute('class')
        action = self.getNewAction(widget, id)

        # if node has alternative label, display it instead
        if a_node.hasAttribute('label'):
            action.setText(a_node.getAttribute('label'))

        if a_node.hasAttribute('checkable'):
            action.setCheckable(bool(a_node.getAttribute('checkable')))

        if a_node.hasAttribute('icon'):
            icon = a_node.getAttribute('icon')
            # TODO
            # action.setIcon(icon)
        return action

    def buildMenu(self, widget, m_node):
        menu = None
        if m_node.hasAttribute('class'):
            klass = self.menus.get(m_node.getAttribute('class'))
            if klass is None:
                return None
            menu = klass(widget)
        else:
            menu = taurusaction.TaurusMenu(widget)
            menu.buildFromXML(m_node)
        return menu

    def createAction(self, parent, text, shortcut=None, icon=None, tip=None,
                     toggled=None, triggered=None, data=None,
                     context=Qt.Qt.WindowShortcut):
        """Create a QAction"""
        action = Qt.QAction(text, parent)
        if triggered is not None:
            action.triggered.connect(triggered)
        if toggled is not None:
            action.toggled.connect(toggled)
            action.setCheckable(True)
        if icon is not None:
            if isinstance(icon, string_types):
                icon = Qt.QIcon.fromTheme(icon)
            action.setIcon(icon)
        if shortcut is not None:
            action.setShortcut(shortcut)
        if tip is not None:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if data is not None:
            action.setData(data)
        # TODO: Hard-code all shortcuts and choose context=Qt.WidgetShortcut
        # (this will avoid calling shortcuts from another dockwidget
        #  since the context thing doesn't work quite well with these widgets)
        action.setShortcutContext(context)
        return action
